/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|          Program Name: Board Support Package for ATM644RS485V11                                                       |
|                Author: Michael Krmer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 09-Jun-2022                                                                                    |
|                  Tool: AVR Studio 7.0                                                                                 |
|                Device: ATMega644PA-AU                                                                                 |
|               Version: 0.1                                                                                            |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Board Support Package                                                                                       |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+

Preprocessor-Macros. Define the following preprocessor symbols to achieve the respective behavior:
====================

SIMULATION:  If SIMULATION is defined, certain hardware waits and optimizations are turned off.
             SIMULATION must be undefined for final hardware.

DEBUGCODE:   If DEBUGCODE is defined, some code segments for debugging will be enabled.
             DEBUGCODE should be undefined for the release version.

BOOTLOADER:  If BOOTLOADER is defined, a reduced version of the BSP will be generated to be used
             for the bootloader code.

Note:        Nested interrupts are not permitted!
=====        It is assumed that an interrupt handler is never interrupted.

ToDo:
=====
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#include "project.h" /* will include all other relevant files */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       EEPROM data                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Inline functions                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#ifdef SIMULATION
__inline static void go_idle(void) {} /* do nothing */
#else
__inline static void go_idle(void) {SMCR = (0<<SM2) | (0<<SM1) | (0<<SM0) | (1<<SE) ;    /* select idle mode and enable sleep mode */
                                   __asm("sleep") ;     /* goto sleep mode */
                                   SMCR = 0 ; }         /* clear sleep mode enable */
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       start_application                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Code to start application from bootloader                                                                             |
| Returns false if write data does not match the magic number. Will not return on success.                              |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool start_application(uint16_t addr, uint16_t *data)
{
   if (*data == MAGIC_NUMBER)
   {
      __asm("cli") ; /* Turn off global interrupt */
      iv2app() ;     /* move interrupt vectors to application space */
      goto *NULL ;   /* Jump to application Reset vector */
   }
   return(false) ; /* mismatch of magic number */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       start_bootloader                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Start bootloader almost unconditionally. May hang if bootloader is not available!                                     |
| Word at destination address is checked for unprogrammed state.                                                        |
| Returns false if write data does not match the magic number. No reply on success.                                     |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool start_bootloader(uint16_t addr, uint16_t *data)
{
   if ((*data == MAGIC_NUMBER) && (pgm_read_word(0xe000) != 0xffff))
   {
      __asm("cli") ; /* Turn off global interrupt */
      goto *0x7000 ; /* Jump to bootloader Reset vector */
   }
   return(false) ; /* mismatch of magic number or destination word is unprogrammed */
}

#ifdef BOOTLOADER
/*
+-----------------------------------------------------------------------------------------------------------------------+
|       prog_enable                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Code to enable flash programming                                                                                      |
| Returns false if write data does not match the magic number.                                                          |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool prog_enable(uint16_t addr, uint16_t *data)
{
   if (*data == MAGIC_NUMBER)
   {
      start_timer(&prog_timer, PROG_TIMEOUT) ;
      return(true) ; /* timer started */
   }
   return(false) ; /* mismatch of magic number */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       read_to                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| This is a callback function for the modbus library. It reads the remaining timeout for flash programming.             |
|                                                                                                                       |
| Returns always true.                                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool read_to(uint16_t addr, uint16_t *data)
{
   *data = prog_timer.timer ;
   return (true) ;
}
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       initial_things                                                                                                  |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Do initialization after a reset occured.                                                                              |
|                                                                                                                       |
| This function is executed immediatelly after main() is started. The CPU is not yet initialized, it runs at its        |
| default speed (1 or 8 MHz). Determine cause of reset and optionally initialize reset- and error counters.             |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void initial_things(void)
{
#ifdef BOOTLOADER
   reset_flags = MCUSR ; /* save Reset flags */
   iv2boot() ; /* move interrupt vectors to boot area */
   wdt_off() ; /* Switch off the watchdog timer. WDT may still be enabled from an intentionally issued WDT reset. */
   noapp_start = false ; /* don't start application if noapp_start is set true during startup delay */
   application_is_healthy = false ; /* will be set true if CRC check succeeds */

#ifndef SIMULATION
      _delay_loop_2 (25000) ;    /* wait 100 ms for power settling (four CPU cycles per iteration) */
#endif

   MCUSR = 0 ; /* Clear reset flags in MCU status register */
   module_address = MODBUS_DEVICEADDR ;

/* get communication timings */
   ubrrval = pgm_read_word(&comm_times[BAUDRATEINDEX].ubrr) ;
   t25 = pgm_read_word(&comm_times[BAUDRATEINDEX].t25) ;
   t35 = pgm_read_word(&comm_times[BAUDRATEINDEX].t35) ;
#else
bool rc ;
   reset_flags = MCUSR ; /* save Reset flags */
   wdt_off() ; /* Switch off the watchdog timer. WDT may still be enabled from an intentionally issued WDT reset. */
   if ((reset_flags & (1<<PORF)) != 0)
   { /* This is a cold start */
      software_reset = false ;
      software_reset_count = 0 ;
      ext_reset_count = 0 ;
      bo_reset_count = 0 ;
      wdt_reset_count = 0 ;
      reset_count = 0 ;
      crc_error_count_l = 0 ;
      crc_error_count_h = 0 ;
      message_count_l = 0 ;
      message_count_h = 0 ;
      total_message_count_l = 0 ;
      total_message_count_h = 0 ;
      com_error_count_l = 0 ;
      com_error_count_h = 0 ;
      com_rx_count_l = 0 ;
      com_rx_count_h = 0 ;
      com_tx_count_l = 0 ;
      com_tx_count_h = 0 ;
      on_time_l = 0 ;
      on_time_m = 0 ;
      on_time_h = 0 ;

      rtc_active   = false ;
      rtc_midnight = false ;
      rtc_cs    = 0 ;
      rtc_sec   = 0 ;
      rtc_min   = 0 ;
      rtc_hour  = 0 ;
      rtc_day   = 0 ;
      rtc_month = 0 ;
      rtc_year  = 0 ;
   }

   reset_count++ ;
   if ((reset_flags & (1<<WDRF)) != 0)
   { /* This is a WDT reset */
      if (software_reset)
      { /* this is an intentional software reset */
         software_reset_count++ ;
         software_reset = false ;
      }
      else wdt_reset_count++ ;
   }
   if ((reset_flags & (1<<BORF)) != 0)  bo_reset_count++ ;       /* This is a brown-out reset */
   if ((reset_flags & (1<<EXTRF)) != 0) ext_reset_count++ ;      /* This is a external reset */

   MCUSR = 0 ; /* Clear reset flags in MCU status register */

/* read master control block from EEPROM */
   rc = mcb_eepread(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0])) ; /* read mcb from EEPROM to RAM */
   if (!rc)
   { /* No valid mcb found. Initialise EEPROM with default values */
      mcb_eepinit(&eemcb[0], &default_mcb) ;
      mcb_pgmread(&mcb, &default_mcb) ; /* get default mcb from program memory just in case EEPROM is fatally ill */
   }

   module_address = mcb.module_address ;
   baudrate_10th = mcb.br10th ;

   if (mcb.rtc_correction == 0) rtc_correction = 0 ;
   else
   {
      rtc_correction_reload = abs(mcb.rtc_correction) ;
      rtc_correction_counter = abs(mcb.rtc_correction) ;
      rtc_correction = (mcb.rtc_correction > 0) ? 1 : -1 ;
   }

/* get communication timings */
   baudrate_index = get_br10th(baudrate_10th) ;
   ubrrval = pgm_read_word(&comm_times[baudrate_index].ubrr) ;
   t25 = pgm_read_word(&comm_times[baudrate_index].t25) ;
   t35 = pgm_read_word(&comm_times[baudrate_index].t35) ;
#endif
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       init_CPU                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Initialize device                                                                                                     |
|                                                                                                                       |
| This function is executed once after a hardware reset. Do all initialization, which survives a power down mode.       |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void init_CPU(void)
{
#ifndef SIMULATION
/* set clock prescaler */
   set_system_clock_prescaler() ;
#endif

/* Switch off unused peripherals for power reduction */
#ifdef USE_ADC
   PRR0 = (1<<PRTWI) | (1<<PRTIM2) | (0<<PRTIM1) | (1<<PRTIM0) | (1<<PRUSART1) | (0<<PRUSART0) | (1<<PRSPI) | (0<<PRADC) ; /* TIMER1, USART0 and ADC remain on */
#else
   ADCSRA = (0<<ADEN) ; /* disable ADC */
   PRR0 = (1<<PRTWI) | (1<<PRTIM2) | (0<<PRTIM1) | (1<<PRTIM0) | (1<<PRUSART1) | (0<<PRUSART0) | (1<<PRSPI) | (1<<PRADC) ; /* TIMER1 and USART0 remain on */
#endif

/* Initialize ports */
/* Port A */
   PORTA = PORTA_PULLUP|0x00 ;  /* pull-ups as required, all outputs low */
   DDRA  = PORTA_DIR ;          /* enable outputs */

/* Port B */
   PORTB = PORTB_PULLUP|0x00 ;  /* pull-ups as required, all outputs low */
   DDRB  = PORTB_DIR ;          /* enable outputs */

/* Port C */
   PORTC = PORTC_PULLUP|0x00 ;  /* pull-ups as required, all outputs low */
   DDRC  = PORTC_DIR ;          /* enable outputs */

/* Port D */
   PORTD = PORTD_PULLUP|0x00 ;  /* pull-ups as required, all outputs low */
   DDRD  = PORTD_DIR ;          /* enable outputs */

   init_UART() ;

#ifndef SIMULATION
// occasional problems during simulation
#ifdef USE_TEMPSENSOR
   I2C_PORTCONFIG() ; /* configure I2C interface */
   i2c_write8(TS0_ADDR,1,0x60) ; /* write config register: 12 bits resolution, continuous conversion  */
#endif
#endif

#ifdef USE_ADC
   /* use AREF for voltage reference, right adjust, single ended ADC0 */
   ADMUX = (0<<REFS1) | (0<<REFS0) | (0<<ADLAR) | (0<<MUX4) | (0<<MUX3) | (0<<MUX2) | (0<<MUX1) | (0<<MUX0) ;

   /* enable ADC, prescaler = 128 (clock freq must be between 50 and 200 kHz, so this setting is good for 6.4~20 MHz CPU clock */
   ADCSRA = (1<<ADEN) | (0<<ADSC) | (0<<ADATE) | (0<<ADIF) | (0<<ADIE) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0) ;

   /* use Free running mode */
   ADCSRB = (0<<ADTS2) | (0<<ADTS1) | (0<<ADTS0) ;

   /* Disable digital PA0 input */
   DIDR0 = (0<<ADC7D) | (0<<ADC6D) | (0<<ADC5D) | (0<<ADC4D) | (0<<ADC3D) | (0<<ADC2D) | (0<<ADC1D) | (1<<ADC0D) ;
#endif

/* Timer 1 is used for 10 ms interrupts */
/* Init 16-bit Timer 1 for 10 ms interrupts */
   TCCR1A = 0x00 ;     /* Normal port operation: OC1A/B disconnected, normal waveform generation mode */
   TCCR1B = ((0<<ICNC1) | (0<<ICES1) | (0<<WGM13) | (0<<WGM12) | PRSC_TCCR1B) ;     /* normal waveform generation mode, clk(I/O) divider */
   TCCR1C = 0x00 ;     /* unused */
   TCNT1 = 0 ;         /* make sure to start at 0 */
   OCR1A = RLD_OCR1A  ; /* load compare register for next interrupt in 10 ms */
   TIFR1 = ((1<<ICF1) | (1<<OCF1B) | (1<<OCF1A) | (1<<TOV1)) ; /* clear any possibly pending interrupt */
   TIMSK1 = (1<<OCIE1A) ; /* enable Output Compare 1A interrupt */

/* disable timers */
#ifdef BOOTLOADER
   prog_timer.active = false ;
   start_timer(&startup_delay, STARTUP_DELAY) ;
#endif
#ifdef USE_ADC
   adc_timer.active = false ;
#endif
#ifdef USE_TEMPSENSOR
   ts_timer.active = false ;
#endif

   __asm("sei") ; /*  Turn on global interrupt */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       wait_for_rx                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Wait for reception of a Modbus command. Do some house keeping while waiting.                                          |
|                                                                                                                       |
| This function executes regularly while waiting for a bus transaction. Interrupts shall not be disabled.               |
| Bus messages may be lost, if execution takes too long.                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void wait_for_rx(RXBUFFER *rxb)
{
#ifndef BOOTLOADER
uint8_t num_days ;
#endif
uint16_t tmp ;

   while ((rxb->status == RXSTATE_RECEIVING) || (rxb->status == RXSTATE_TIMEOUT))
   {
      __asm("wdr") ; /* say hello to the watchdog */

#ifdef USE_TEMPSENSOR
      if (!ts_timer.active)
      { /* read temperature every now and then */
         i2c_setptr(TS0_ADDR, 0) ;
         temperature = ((int16_t) (i2c_read16(TS0_ADDR) & 0xfff0)) / 16 ;
         start_timer(&ts_timer, TS_READ_INTERVAL) ;
      }
#endif

#ifdef USE_ADC
      if (!adc_timer.active)
      { /* read VIN voltage via ADC every now and then */
         if ((ADCSRA & (1<<ADSC)) == 0) /* no conversion ongoing */
            vin_voltage = ADC ; /* read last conversion result, first one will be wrong  */
         /* start next conversion */
         ADCSRA |= (1<<ADSC) ;
         start_timer(&adc_timer, ADC_READ_INTERVAL) ;
      }
#endif

      wait_for_rx_user() ; /* Execute user specific functions */

/* do some house keeping */
#ifndef BOOTLOADER
      if (rtc_midnight)
      { /* update date */
         rtc_midnight = false ;
         /* days per month: 28: Feb; 29: Feb (leap year); 30: Apr, Jun, Sept, Nov; 31: Jan, Mar, May, Jul, Aug, Oct, Dec */
         if (rtc_month == 2)
         { /* it is February, check for leap year. Note that year offset is 2000 */
            num_days = 28 ;
            if (((rtc_year % 4) == 0) && (rtc_year != 100) && (rtc_year != 200)) num_days = 29 ; /* this is a leap year */
         }
         else
         {
            if ((rtc_month == 4) || (rtc_month == 6) || (rtc_month == 9) || (rtc_month == 11)) num_days = 30 ; /* 30 days per month */
            else num_days = 31 ; /* this month has 31 days */
         }
         if (rtc_day < num_days) rtc_day++ ;
         else
         { /* new month */
            rtc_day = 1 ;
            if (rtc_month < 12) rtc_month++ ;
            else
            { /* new year */
               rtc_month = 1 ;
               rtc_year++ ;
            }
         }
      }
#endif
      go_idle() ; /* idle until next interrupt, i.e. max. 10 ms. No need to switch off watchdog */
   }

/* data received. Add number of received bytes to total count */
   tmp = com_rx_count_l + rxb->bytecounter ;
   if (tmp < com_rx_count_l) com_rx_count_h++ ; /* we had an overflow in the lower half */
   com_rx_count_l = tmp ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       wait_for_tx                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Wait for transmit completion                                                                                          |
|                                                                                                                       |
| This function executes regularly while waiting for a bus transaction. Interrupts shall not be disabled.               |
| Bus messages may be lost, if execution takes too long.                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void wait_for_tx(TXBUFFER *txb)
{
uint16_t tmp ;
   while (!txb->completed)
   { /* wait for completion */
      __asm("wdr") ; /* say hello to the watchdog */
      _delay_loop_1 (10) ; /* wait a little while (three CPU cycles per iteration) */
   }

/* all data sent. Add number of sent bytes to total count */
   tmp = com_tx_count_l + txb->bytecounter ;
   if (tmp < com_tx_count_l) com_tx_count_h++ ; /* we had an overflow in the lower half */
   com_tx_count_l = tmp ;
}

#ifndef BOOTLOADER
/*
+-----------------------------------------------------------------------------------------------------------------------+
|       restart                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Code to restart application by software reset (WDT reset)                                                             |
| Returns false if write data does not match the magic number. No reply on success.                                     |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool restart(uint16_t addr, uint16_t *data)
{
   if (*data == MAGIC_NUMBER)
   {
      software_reset = true ;
      issue_wdt_reset() ; /* Restart module by generating a WDT reset */
   }
   return(false) ; /* mismatch of magic number */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       mcb_eepread                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read currently active mcb from EEPROM into RAM                                                                        |
| "eemcb" points to the first mcb in the EEPROM (source address).                                                       |
| "mcb" points to the mcb in RAM (target address).                                                                      |
| "num_blocks" is the size of the mcb array in EEPROM.                                                                  |
|                                                                                                                       |
| Return true if valid block was read, otherwise return false.                                                          |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool mcb_eepread(MCB *eemcb, MCB *mcb, uint8_t num_blocks)
{
uint8_t i, j ;
uint8_t val ;
uint8_t *src ;
uint8_t *target ;
uint16_t crc ;
bool rc ;
   rc = false ;
   for (i=0; i<num_blocks; i++)
   {
      val = eeprom_read_byte(&eemcb[i].valid) ;
      if (val == VALID)
      { /* valid entry found, retrieve data */
         src = (uint8_t *) &eemcb[i] ;
         target = (uint8_t *) mcb ;
         for (j=0; j<sizeof(MCB); j++) target[j] = eeprom_read_byte(&src[j]) ;
         crc = modbus_crc16(target, sizeof(MCB)-2) ; /* generate CRC */
         rc = (crc == mcb->crc) ;
      }
   }
   return(rc) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       mcb_eepwrite                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Write mcb from RAM into new mcb in EEPROM. Increment number of write cycles.                                          |
| "eemcb" points to the first mcb in the EEPROM (target address).                                                       |
| "mcb" points to the mcb in RAM (source address).                                                                      |
| "num_blocks" is the size of the mcb array in EEPROM.                                                                  |
|                                                                                                                       |
| Return true if block was successfully written, otherwise return false.                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool mcb_eepwrite(MCB *eemcb, MCB *mcb, uint8_t num_blocks)
{
uint8_t i ;
uint8_t val ;
uint8_t *src ;
uint8_t *oldmcb ;
uint8_t *newmcb ;
uint16_t crc ;
uint16_t countl ;
uint16_t counth ;
bool rc ;
   rc = false ;
   if (valid_baudrate(mcb->br10th))
   { /* ignore if invalid baudrate selected */
      for (i=0; ((i<num_blocks) && !rc); i++)
      { /* find first valid block */
         val = eeprom_read_byte(&eemcb[i].valid) ;
         if (val == VALID)
         { /* valid block found */
            rc = true ;
            oldmcb = (uint8_t *) &eemcb[i] ; /* save its address */
            countl = eeprom_read_word(&eemcb[i].pgm_cntrl) ;
            counth = eeprom_read_word(&eemcb[i].pgm_cntrh) ;
            if (i < (num_blocks-1))
               newmcb = (uint8_t *) &eemcb[i+1] ; /* pointer to next mcb */
            else
               newmcb = (uint8_t *) &eemcb[0] ; /* wrap over to first mcb */
         }
      }
      if (rc)
      { /* old and new address found. Increment number of write cycles, generate CRC and write data */
         mcb->valid = VALID ; /* just in case... */
         countl++ ;
         if (countl == 0) counth++ ;
         mcb->pgm_cntrl = countl ;
         mcb->pgm_cntrh = counth ;
         src = (uint8_t *) mcb ;
         crc = modbus_crc16(src, sizeof(MCB)-2) ; /* generate CRC */
         mcb->crc = crc ;
         for (i=1; i<sizeof(MCB); i++) eeprom_write_byte(&newmcb[i], src[i]) ;
         eeprom_write_byte(&newmcb[0], src[0]) ; /* stamp the new mcb valid */
         eeprom_write_byte(&oldmcb[0], ~VALID) ; /* finally stamp the old mcb invalid */
      }
   }
   else mcb_eepread(eemcb, mcb, num_blocks) ; /* restore original mcb */
   return(rc) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       mcb_eepinit                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Initialise mcb in EEPROM with default values from program memory.                                                     |
| "eemcb" points to the first mcb in the EEPROM. Always only the first mcb will be initialised.                         |
| "pgmmcb" points to the mcb in program memory, which holds the default values.                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void mcb_eepinit(MCB *eemcb, const MCB *pgmmcb)
{
static volatile MCB tmpmcb ;
uint8_t *src ;
uint8_t *target ;
uint16_t crc ;
uint8_t i ;
   src = (uint8_t *) pgmmcb ;
   target = (uint8_t *) &tmpmcb ;
   for (i=0; i<sizeof(MCB); i++) target[i] = pgm_read_byte(&src[i]) ; /* get data from progmem to local buffer */
   crc = modbus_crc16(target, sizeof(MCB)-2) ; /* generate CRC */
   tmpmcb.crc = crc ;
   src = (uint8_t *) &tmpmcb ;
   target = (uint8_t *) &eemcb[0] ; /* target address is block 0 */
   for (i=1; i<sizeof(MCB); i++) eeprom_write_byte(&target[i], src[i]) ; /* write block into EEPROM */
   eeprom_write_byte(&target[0], src[0]) ; /* finally stamp the mcb valid */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       mcb_pgmread                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read default mcb from program memory to ram.                                                                          |
| "mcb" points to the mcb in RAM (target address).                                                                      |
| "pgmmcb" points to the mcb in program memory, which holds the default values.                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void mcb_pgmread(MCB *mcb, const MCB *pgmmcb)
{
uint8_t *src ;
uint8_t *target ;
uint16_t crc ;
uint8_t i ;
   src = (uint8_t *) pgmmcb ;
   target = (uint8_t *) mcb ;
   for (i=0; i<sizeof(MCB); i++) target[i] = pgm_read_byte(&src[i]) ; /* get data from progmem to ram buffer */
   crc = modbus_crc16(target, sizeof(MCB)-2) ; /* generate CRC */
   mcb->crc = crc ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       read_mcb                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| This is a callback function for the modbus library. It reads the EEPROM MCB into the RAM mcb.                         |
|                                                                                                                       |
| Return true if valid block was read, otherwise return false.                                                          |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool read_mcb(uint16_t addr, uint16_t *data)
{ /* read mcb from EEPROM to RAM */
   return (mcb_eepread(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0]))) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       write_mcb                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| This is a callback function for the modbus library. It writes the RAM mcb back to the EEPROM mcb.                     |
|                                                                                                                       |
| Return true if block was successfully written, otherwise return false.                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool write_mcb(uint16_t addr, uint16_t *data)
{ /* write mcb from RAM to EEPROM */
   return (mcb_eepwrite(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0]))) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       bool valid_baudrate(uint16_t br)                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Verify that baudrate is found in the list of supported_baudrates.                                                     |
|                                                                                                                       |
| Return true if supported, false otherwise.                                                                            |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool valid_baudrate(uint16_t br)
{
bool supported ;
uint8_t i ;
   supported = false ;
   i = 0 ;
   while ((!supported) && (i<CT_SIZE) && (supported_baudrates[i] != 0))
   {
      supported = (supported_baudrates[i] == br) ;
      i++ ;
   }
   return (supported) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       get_supported_baudrates()                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Fill register table with supported baud rates.                                                                        |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void get_supported_baudrates(void)
{
uint8_t i, j ;
   for (i=0;i<CT_SIZE;i++) supported_baudrates[i] = 0 ;
   j = 0 ;
   for (i=0;i<CT_SIZE;i++)
   {
      if (pgm_read_word(&comm_times[i].ubrr) != 0)
      {
         supported_baudrates[j] = pgm_read_word(&comm_times[i].br10) ;
         j++ ;
      }
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       uint8_t get_br10th(uint16_t br10th)                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Get index into comm_times table for specified baudrate. Return index for BAUDRATE10TH if not found.                   |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint8_t get_br10th(uint16_t br10th)
{
uint8_t i ;
uint8_t index, index_default ;
   index = index_default = 255 ;
   for (i=0;i<CT_SIZE;i++)
   {
      if (pgm_read_word(&comm_times[i].br10) == BAUDRATE10TH) index_default = i ;
      if ((pgm_read_word(&comm_times[i].ubrr) != 0) && (pgm_read_word(&comm_times[i].br10) == br10th)) index = i ;
   }
   if (index == 255) index = index_default ; /* default baudrate if not found */
   if (index == 255) index = 0 ; /* should not happen, default baudrate should always be found */
   return(index) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       clear_counters                                                                                                  |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Clear various counters                                                                                                |
|                                                                                                                       |
| Returns true if no error occured, false otherwise.                                                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool clear_counters(uint16_t addr, uint16_t *data)
{
   if (*data == MAGIC_NUMBER)
   {
      software_reset_count = 0 ;
      ext_reset_count = 0 ;
      bo_reset_count = 0 ;
      wdt_reset_count = 0 ;
      reset_count = 0 ;
      total_message_count_l = 0 ;
      total_message_count_h = 0 ;
      message_count_l = 0 ;
      message_count_h = 0 ;
      crc_error_count_l = 0 ;
      crc_error_count_h = 0 ;
      com_error_count_l = 0 ;
      com_error_count_h = 0 ;
      com_rx_count_l = 0 ;
      com_rx_count_h = 0 ;
      com_tx_count_l = 0 ;
      com_tx_count_h = 0 ;
      return(true) ;
   }
   else return(false) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       reset_deviceid                                                                                                  |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Code to reset the device ID                                                                                           |
|                                                                                                                       |
| Returns true if no error occured, false otherwise.                                                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool reset_deviceid(uint16_t addr, uint16_t *data)
{
   if (*data == MAGIC_NUMBER)
   { /* MCB is about to be updated, read current MCB from EEPROM */
      if (mcb_eepread(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0]))) /* read mcb from EEPROM to RAM */
      { /* Valid mcb found. */
         mcb.module_address = MODBUS_DEVICEADDR ;
         mcb_eepwrite(&eemcb[0], &mcb, sizeof(eemcb)/sizeof(eemcb[0])) ; /* write mcb from RAM to EEPROM, ignore any possible error */
      }
      return(true) ;
   }
   else return(false) ;
}
#endif

#ifdef USE_TEMPSENSOR

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_read8                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read data from an I2C cannel                                                                                          |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint8_t i2c_read8(uint8_t addr)
{
uint8_t data ;
bool ack ;

   data = 0 ;
   I2C_START() ;
   ack = i2c_sendbyte(addr<<1 | 0x01) ; /* transfer address and R/W = high (for read) */
   if (ack) data  = i2c_receivebyte(true) ; /* read data byte */
   I2C_STOP() ;

   return(data) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_read16                                                                                                      |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read data from an I2C channel                                                                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint16_t i2c_read16(uint8_t addr)
{
uint16_t data ;
bool ack ;

   data = 0 ;
   I2C_START() ;
   ack = i2c_sendbyte(addr<<1 | 0x01) ; /* transfer address and R/W = high (for read) */
   if (ack)
   {
      data  = ((uint16_t) i2c_receivebyte(false)) << 8 ; /* read MSB */
      data |=  (uint16_t) i2c_receivebyte(true) ; /* read LSB */
   }
   I2C_STOP() ;

   return(data) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_write8                                                                                                      |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Write 8-bit data to an I2C channel                                                                                    |
|                                                                                                                       |
| Returns true if successful (all ACK properly received)                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool i2c_write8(uint8_t addr, uint8_t pointer, uint8_t data)
{
uint8_t ack ;

   I2C_START() ;
   ack  = i2c_sendbyte(addr<<1) ; /* transfer address and R/W = low (for write) */
   if (ack) ack |= i2c_sendbyte(pointer) ; /* transfer register pointer */
   if (ack) ack |= i2c_sendbyte(data) ; /* transfer byte */
   I2C_STOP() ;

   return(ack == 0) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_write16                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Write 16-bit data to an I2C channel                                                                                   |
|                                                                                                                       |
| Returns true if successful (all ACK properly received)                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool i2c_write16(uint8_t addr, uint8_t pointer, uint16_t data)
{
uint8_t ack ;

   I2C_START() ;
   ack  = i2c_sendbyte(addr<<1) ; /* transfer address and R/W = low (for write) */
   if (ack) ack |= i2c_sendbyte(pointer) ; /* transfer register pointer */
   if (ack) ack |= i2c_sendbyte(data>>8) ; /* transfer MSB */
   if (ack) ack |= i2c_sendbyte(data&0xff) ; /* transfer LSB */
   I2C_STOP() ;

   return(ack == 0) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_setptr                                                                                                      |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Set register pointer for I2C                                                                                          |
|                                                                                                                       |
| Returns true if successful (all ACK properly received)                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool i2c_setptr(uint8_t addr, uint8_t pointer)
{
uint8_t ack ;

   I2C_START() ;

   ack  = i2c_sendbyte(addr<<1) ; /* transfer address and R/W = low (for write) */
   if (ack) ack |= i2c_sendbyte(pointer) ; /* transfer pointer */

   I2C_STOP() ;

   return(ack == 0) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_sendbyte                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Write byte data and read acknowledge                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool i2c_sendbyte(uint8_t data)
{
uint8_t ack ;
int i ;
   for (i=7; i>=0; i--)
   { /* send bits 7..0 (high order bit first) */
      I2C_SDASET(data>>i) ; /* set SDA */
      I2C_SCLHIGH() ;
      I2C_SCLLOW() ;
   }
   I2C_SDAHIGH() ; /* release pull down to read ack */
   I2C_SCLHIGH() ;
   ack = I2C_SDAGET() ; /* read acknowledge from slave */
//   I2C_SDALOW() ; /* pull SDA down again */
   I2C_SCLLOW() ;

   return(ack == 0) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       i2c_receivebyte                                                                                                 |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Receive byte data and read acknowledge                                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint8_t i2c_receivebyte(bool lastbyte)
{
int i ;
uint8_t data ;
   data = 0 ;
   I2C_SDAHIGH() ; /* release pull down */
   for (i=0; i<8; i++)
   { /* get byte */
      I2C_SCLHIGH() ;
      data = (data << 1) | I2C_SDAGET() ;
      I2C_SCLLOW() ;
   }
   if (lastbyte)
      I2C_SDAHIGH() ; /* nack (high) */
   else
      I2C_SDALOW() ; /* ack (low) */
   I2C_SCLHIGH() ;
   I2C_SCLLOW() ;
   I2C_SDAHIGH() ; /* release pull down */

   return(data) ;
}

#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       TIMER1_COMPA                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Interrupt handler for timer1 compare match A. 10 ms interrupt.                                                        |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
ISR(TIMER1_COMPA_vect)
{
   OCR1A += RLD_OCR1A ; /* prepare for next interrupt in 10 ms from this interrupt */

#ifdef USE_TEMPSENSOR
   if (ts_timer.active)
   {
      if (ts_timer.timer != 0) ts_timer.timer-- ;
      if (ts_timer.timer == 0) ts_timer.active = false ;
   }
#endif

#ifdef USE_ADC
   if (adc_timer.active)
   {
      if (adc_timer.timer != 0) adc_timer.timer-- ;
      if (adc_timer.timer == 0) adc_timer.active = false ;
   }
#endif

#ifdef BOOTLOADER
   if (prog_timer.active)
   {
      if (prog_timer.timer != 0) prog_timer.timer-- ;
      if (prog_timer.timer == 0) prog_timer.active = false ;
   }

   if (startup_delay.active)
   {
      if (startup_delay.timer != 0) startup_delay.timer-- ;
      if (startup_delay.timer == 0) startup_delay.active = false ;
   }

#endif

   user_int10ms() ; /* do any user specific 10ms handling */

#ifndef BOOTLOADER
/* increment power-on time */
   on_time_l++ ;
   if (on_time_l == 0)
   {
      on_time_m++ ;
      if (on_time_m == 0) on_time_h++ ;
   }

/* Real Time Clock */
   if (rtc_active)
   {
      if (rtc_correction != 0)
      {
         if (rtc_cs == 60) inhibit_rtc_correction = false ;
         if ((!inhibit_rtc_correction) && (rtc_cs == 50))
         {
            rtc_correction_counter-- ;
            if (rtc_correction_counter == 0)
            { /* apply correction */
               rtc_correction_counter = rtc_correction_reload ;
               if (rtc_correction > 0) rtc_cs++ ;
               if (rtc_correction < 0)
               {
                  rtc_cs-- ;
                  inhibit_rtc_correction = true ; /* prevents double correction if rtc_cs was decremented previously */
               }
            }
         }
      }

      if (rtc_cs == 99)
      { /* next second */
         rtc_cs = 0 ;
         if (rtc_sec == 59)
         {
            rtc_sec = 0 ;
            if (rtc_min == 59)
            {
               rtc_min = 0 ;
               if (rtc_hour == 23)
               {
                  rtc_hour = 0 ;
                  rtc_midnight = true ;
               }
               else rtc_hour++ ;
            }
            else rtc_min++ ;
         }
         else rtc_sec++ ;
      }
      else rtc_cs++ ;
   }
#endif
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       TIMER1_COMPB                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Interrupt handler for communication timeouts.                                                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
ISR(TIMER1_COMPB_vect)
{
   switch(prxb->status)
   {
      case RXSTATE_RECEIVING:
         prxb->status = RXSTATE_TIMEOUT ;
         OCR1B = TCNT1+(t35-t25) ;  /* set frame timeout, t25 is already passed */
      break ;

      case RXSTATE_TIMEOUT:
         prxb->status = RXSTATE_COMPLETED ; /* we had a character timeout before, so this is the end of the frame */
         TOINT_disable() ; /* disable further timeout interrupts */
      break ;

      default:
         TOINT_disable() ; /* disable further timeout interrupts */
      break ;
   }
}


