﻿/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|             File Name: BSP_ATM644RS485V11.h                                                                           |
|                Author: Michael Krämer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 07.06.2022                                                                                     |
|                  Tool: AVR Studio 7.0 and Visual Studio                                                               |
|                Device: ATMEGA644                                                                                      |
|               Version: 0.1                                                                                            |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Include file for ATM644RS485V11 Board Support Package                                                       |
|           This file is also included by the host compiler                                                             |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#ifndef BSP_ATM644RS485V11

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define BSP_ATM644RS485V11 1
#define MAGIC_NUMBER 0x39f1     /* magic number to restart remote device */
#define EEP_SIZEW (8u * 128u) /* size of EEPROM in words (2 bytes) */
#define FLASH_SIZEW (32u * 1024u) /* size of program memory in words (2 bytes) */

#define PROG_TIMEOUT 1000    /* Flash programming enable timeout in cs */
#define STARTUP_DELAY 200    /* Startup delay in cs */

#define APP_PAGECNT 224u                /* Number of pages for application code (ATMega644PA-AU) */
#define BOOT_PAGECNT 32u                /* Number of pages for bootloader code (ATMega644PA-AU) */
#define PAGESIZEW 128u                  /* page size of flash memory in 16-bit words */
#define PAGESIZEB (PAGESIZEW*2u)        /* page size of flash memory in bytes */
#define CHUNKSIZEW (PAGESIZEW/4u)       /* number of 16-bit words which is transmitted in one transfer */
#define CHUNKSIZEB (CHUNKSIZEW*2u)      /* number of bytes which is transmitted in one transfer */
#define EEPCHUNKSIZEW 32u               /* number of 16-bit words which is transmitted in one transfer for EEPROM programming */
#define SIGNATUREPAGE (APP_PAGECNT-1u)  /* last page is used for signature */

#ifndef WIN32
#ifdef USE_TEMPSENSOR
/*
Timing requirements (of TMP275 in fast mode):
max clock speed: 400 kHz
bus free time between STOP and START: min 1300 ns
Data hold time after SCL low: 4 .. 900 ns
==> ensured by SCL low period
Data setup time between setting SDA and SCL high: 100 ns (min)
==> insert just one NOP delay if clock speed is higher than 10 MHz
Clock low period:  1300 ns (min)
CLOCK high period:  600 ns (min)
SDA must be stable while SCL is high
*/

#define I2C_SCL (1<<PB3)        /* PB3 is I2C SCL signal */
#define I2C_SDA (1<<PB4)        /* PB4 is I2C SDA signal */
#define I2C_SDAPOS PB4          /* PB4 is I2C SDA signal */
#define TS0_ADDR 0x48           /* I2C address of temperature sensor 0 */

uint8_t i2c_read8(uint8_t addr) ;
uint16_t i2c_read16(uint8_t addr) ;
bool i2c_write8(uint8_t addr, uint8_t pointer, uint8_t data) ;
bool i2c_write16(uint8_t addr, uint8_t pointer, uint16_t data) ;
bool i2c_setptr(uint8_t addr, uint8_t pointer) ;

/* I2C interface functions */
__inline static void I2C_PORTCONFIG(void)
                          { /* Configure ports for I2C channel (PB3 for SCL, PB4 for SDA) */
                             PORTB |=  I2C_SCL ; /* SCL is high when idle */
                             PORTB &= ~I2C_SDA ; /* SDA is low when output is enabled */
                             DDRB |=   I2C_SCL ; /* SCL is output */
                             DDRB &=  ~I2C_SDA ; /* SDA is input for now */
                          }

__inline static void I2C_SU_TIME(void)  {if (CPU_CLK > 10000) _NOP() ;} /* ensure setup time */
__inline static void I2C_SCL_LOW(void)  {_delay_loop_1(1+(1300/(3*NS_PER_CPUCLK))) ;} /* ensure min SCL low time */
__inline static void I2C_SCL_HIGH(void) {_delay_loop_1(1+(600/(3*NS_PER_CPUCLK))) ;} /* ensure min SCL high time */
__inline static void I2C_BUSFREE(void)  {_delay_loop_1(1+(1300/(3*NS_PER_CPUCLK))) ;} /* bus free time between stop and start (I2C Spec: 1300ns min) */

__inline static void I2C_SDALOW(void)  {DDRB  |=  I2C_SDA ; I2C_SU_TIME() ;} /* SDA low */
__inline static void I2C_SDAHIGH(void) {DDRB  &= ~I2C_SDA ; I2C_SU_TIME() ;} /* SDA high */
__inline static void I2C_SDASET(uint8_t data) {DDRB = (DDRB & ~I2C_SDA) | ((~data & 0x01)<<I2C_SDAPOS) ; I2C_SU_TIME() ;} /* SDA high */
__inline static void I2C_SCLLOW(void)  {PORTB &= ~I2C_SCL ; _delay_loop_1(1+(1300/(3*NS_PER_CPUCLK))) ;} /* SCL low */
__inline static void I2C_SCLHIGH(void) {PORTB |=  I2C_SCL ; _delay_loop_1(1+(600/(3*NS_PER_CPUCLK))) ;} /* SCL high */

__inline static uint8_t I2C_SDAGET(void) {return ((PINB & I2C_SDA)>>I2C_SDAPOS) ;} /* Read SDA and move to bit 0 */
__inline static void I2C_START(void) {I2C_SDALOW() ; I2C_SCLLOW() ;} /* generate I2C start condition */
__inline static void I2C_STOP(void)  {I2C_SDALOW() ; I2C_SCLHIGH() ; I2C_SDAHIGH() ; I2C_BUSFREE() ;} /* generate I2C stop condition */

#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void initial_things(void) ;
void init_CPU(void) ;
void wait_for_rx(RXBUFFER *rxb) ;
void wait_for_tx(TXBUFFER *txb) ;
bool restart(uint16_t addr, uint16_t *data) ;
bool start_application(uint16_t addr, uint16_t *data) ;
bool start_bootloader(uint16_t addr, uint16_t *data) ;
bool prog_enable(uint16_t addr, uint16_t *data) ;
bool read_to(uint16_t addr, uint16_t *data) ;
extern void iv2boot(void) ;
extern void iv2app(void) ;
extern void wdt_off(void) ;
extern void wdt_on(void) ;
extern void issue_wdt_reset(void) ;
extern void set_system_clock_prescaler(void) ;
#ifndef BOOTLOADER
bool mcb_eepread(MCB *eemcb, MCB *mcb, uint8_t num_blocks) ;
bool mcb_eepwrite(MCB *eemcb, MCB *mcb, uint8_t num_blocks) ;
void mcb_eepinit(MCB *eemcb, const MCB *pgmmcb) ;
void mcb_pgmread(MCB *mcb, const MCB *pgmmcb) ;
bool read_mcb(uint16_t addr, uint16_t *data) ;
bool write_mcb(uint16_t addr, uint16_t *data) ;
bool valid_baudrate(uint16_t br) ;
void get_supported_baudrates(void) ;
uint8_t get_br10th(uint16_t br10th) ;
bool clear_counters(uint16_t addr, uint16_t *data) ;
bool reset_deviceid(uint16_t addr, uint16_t *data) ;
#endif
#ifdef USE_TEMPSENSOR
bool i2c_sendbyte(uint8_t data) ;
uint8_t i2c_receivebyte(bool lastbyte) ;
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Inline functions                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/
__inline static void TOINT_enable(void)  {TIFR1=(1<<OCF1B); TIMSK1 |=  (1<<OCIE1B) ;} /* enable Output Compare 1B interrupt */
__inline static void TOINT_disable(void) {TIMSK1 &= ~(1<<OCIE1B) ;} /* disable Output Compare 1B interrupt */
__inline static void start_timer(TIMER *timer, uint16_t to)  /* set timeout and activate timer */
                                {timer->active = false ;     /* this access is atomic, DI is not required */
                                 timer->timer = to;
                                 timer->active = true ;}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/
_EXTERN_ uint8_t reset_flags __attribute__ ((section (".noinit"))) ; /* Copy of reset flags from MCU status register */

#ifdef USE_TEMPSENSOR
_EXTERN_ int16_t temperature ; /* current temperature */
#endif

#ifdef BOOTLOADER
_EXTERN_ uint8_t page_buffer[PAGESIZEB] ;  /* page buffer for writing flash memory */
_EXTERN_ bool volatile noapp_start ;   /* Don't start application from bootloader if value is not zero */
_EXTERN_ bool volatile application_is_healthy ;   /* True if application code is healthy */
_EXTERN_ TIMER prog_timer ;   /* timer for flash programming timeout */
_EXTERN_ TIMER startup_delay ;   /* timer for startup delay */
#else
_EXTERN_ MCB mcb ; /* runtime copy of MCB from EEPROM */

/* runtime copies from EEPROM */
_EXTERN_ uint8_t baudrate_index ; /* baudrate, index into "comm_times" table */
_EXTERN_ uint16_t baudrate_10th ; /* baudrate / 10 */

/* NOINIT: the following variables survive the startup code */
/* Module status and error counters. May be read and written via ModBus functions */
_EXTERN_ bool software_reset __attribute__ ((section (".noinit"))) ; /* flag to indicate an intentional software reset */
_EXTERN_ uint16_t software_reset_count __attribute__ ((section (".noinit"))) ; /* Number of software resets since last power-on reset */
_EXTERN_ uint16_t wdt_reset_count __attribute__ ((section (".noinit"))) ; /* Number of watchdog timer resets since last power-on reset */
_EXTERN_ uint16_t reset_count __attribute__ ((section (".noinit"))) ; /* Number of resets since last power-on reset */
_EXTERN_ uint16_t ext_reset_count __attribute__ ((section (".noinit"))) ; /* Number of external resets since last power-on reset */
_EXTERN_ uint16_t bo_reset_count __attribute__ ((section (".noinit"))) ; /* Number of brown-out resets since last power-on reset */
_EXTERN_ uint16_t volatile on_time_l __attribute__ ((section (".noinit"))) ; /* Power-on time in centi-seconds (lower) */
_EXTERN_ uint16_t volatile on_time_m __attribute__ ((section (".noinit"))) ; /* Power-on time in centi-seconds (middle) */
_EXTERN_ uint16_t volatile on_time_h __attribute__ ((section (".noinit"))) ; /* Power-on time in centi-seconds (higher) */

/* RTC counters */
_EXTERN_ volatile bool rtc_active   __attribute__ ((section (".noinit"))) ; /* RTC is active if true */
_EXTERN_ volatile bool rtc_midnight __attribute__ ((section (".noinit"))) ; /* Midnight */
_EXTERN_ volatile uint8_t rtc_cs    __attribute__ ((section (".noinit"))) ; /* cs counter (0..99) */
_EXTERN_ volatile uint8_t rtc_sec   __attribute__ ((section (".noinit"))) ; /* seconds counter (0..59) */
_EXTERN_ volatile uint8_t rtc_min   __attribute__ ((section (".noinit"))) ; /* minutes counter (0..59) */
_EXTERN_ volatile uint8_t rtc_hour  __attribute__ ((section (".noinit"))) ; /* hours counter (0..23) */
_EXTERN_ uint8_t rtc_day   __attribute__ ((section (".noinit"))) ; /* days counter (1..31) */
_EXTERN_ uint8_t rtc_month __attribute__ ((section (".noinit"))) ; /* months counter (1..12) */
_EXTERN_ uint8_t rtc_year  __attribute__ ((section (".noinit"))) ; /* years counter (0..255 = 2000..2255) */

/* RTC is corrected by an extra increment or decrement to the 10ms-counter (rtc_cs) every
   rtc_correction_counter seconds. */
_EXTERN_ volatile uint16_t rtc_correction_counter ; /* down counter of seconds for RTC correction */
_EXTERN_ uint16_t rtc_correction_reload ; /* reload value for expired rtc_correction_counter */
_EXTERN_ int8_t rtc_correction ; /* RTC is corrected forward if positive, backward if negative, no correction if zero */
_EXTERN_ volatile bool inhibit_rtc_correction ; /* temporarily inhibits RTC correction to prevent double correction */
_EXTERN_ uint16_t vin_voltage ; /* supply voltage */

_EXTERN_ uint16_t supported_baudrates[sizeof(comm_times)/sizeof(comm_times[0])] ;
#endif

#endif /* WIN32 */
#endif /* BSP_ATM644RS485V11 */

