/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|          Program Name: General purpose assembly functions                                                             |
|                Author: Michael Krmer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 14-Jul-2022                                                                                    |
|                  Tool: AVR Studio 7.0                                                                                 |
|                Device: ATMega644PA-AU                                                                                 |
|               Version: 0.1                                                                                            |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Low level functions for Flash self-programming and other general purposes                                   |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+

Notes:
    AVR register conventions (from internet):
    R0: is used as scratch register that need not to be restored after its usage.
    R1: always contains zero
    R18R27, R30, R31: may be destroyed by a function call.
    R2R17, R28, R29: a function that uses such a registers must restore its original content

    Calling Convention:

        An argument is passed either completely in registers or completely in memory.

        To find the register where a function argument is passed, initialize the register number Rn with R26 and follow this procedure:
            If the argument size is an odd number of bytes, round up the size to the next even number.

            Subtract the rounded size from the register number Rn.

            If the new Rn is at least R8 and the size of the object is non-zero, then the low-byte of the argument is passed in Rn. Subsequent bytes of the argument
            are passed in the subsequent registers, i.e. in increasing register numbers.

            If the new register number Rn is smaller than R8 or the size of the argument is zero, the argument will be passed in memory.
            If the current argument is passed in memory, stop the procedure: All subsequent arguments will also be passed in memory.
            If there are arguments left, goto 1. and proceed with the next argument.
        Return values with a size of 1 byte up to and including a size of 8 bytes will be returned in registers. Return values whose size is outside that range will
        be returned in memory.
        If a return value cannot be returned in registers, the caller will allocate stack space and pass the address as implicit first pointer argument to the callee.
        The callee will put the return value into the space provided by the caller.
        If the return value of a function is returned in registers, the same registers are used as if the value was the first parameter of a non-varargs function.
        For example, an 8-bit value is returned in R24 and an 32-bit value is returned R22...R25.
        Arguments of varargs functions are passed on the stack. This applies even to the named arguments.

    For example, suppose a function with the following prototype:

        int func (char a, long b);

    then

        a will be passed in R24.
        b will be passed in R20, R21, R22 and R23 with the LSB in R20 and the MSB in R23.
        The result is returned in R24 (LSB) and R25 (MSB).

    X-Register = r26/r27
    Y-Register = r28/r29
    Z-Register = r30/r31

*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#include <avr/io.h>

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define PAGESIZEW 128           /* page size of flash memory in 16-bit words */
#define PAGESIZEB (PAGESIZEW*2) /* page size of flash memory in bytes */

/* register definitions for subsequent code */
temp1 = 18
temp2 = 19
spmcrval = 20
looplo = 24
loophi = 25

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void wdt_off(void)                                                                                              |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Switch watchdog timer off                                                                                             |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
.global wdt_off
wdt_off:        in r0, _SFR_IO_ADDR(SREG)               ; save SREG
                push r0                                 ; save SREG on stack
                cli                                     ; disable interrupts

                wdr                                     ; reset watchdog timer

                in      temp1,_SFR_IO_ADDR(MCUSR)
                andi    temp1,~(1<<WDRF)
                out     _SFR_IO_ADDR(MCUSR),temp1       ; clear WDRF in MCUSR

                lds     temp1,WDTCSR
                ori     temp1,((1<<WDCE)|(1<<WDE))      ; set WDCE and WDE, keep prescaler setting
                sts     WDTCSR,temp1

                andi    temp1,~(1<<WDE)                 ; clear WDE, keep others
                sts     WDTCSR,temp1                    ; turn off WDT within four instruction cycles

                pop r0                                  ; restore previous SREG
                out _SFR_IO_ADDR(SREG), r0              ; restore SREG and interrupt status
                ret

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void wdt_on(void)                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Switch watchdog timer on                                                                                              |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
.global wdt_on
wdt_on:         cli                                     ; turn off global interrupts
                wdr                                     ; WDT might be on already
                in      temp1,_SFR_IO_ADDR(MCUSR)       ; clear WDRF in MCUSR
                andi    temp1,~(1<<WDRF)
                out     _SFR_IO_ADDR(MCUSR),temp1

                ldi     temp1,((1<<WDCE)|(1<<WDE))      ; set WDCE and WDE
                sts     WDTCSR,temp1

//                ldi     temp1,((0<<WDCE)|(1<<WDE)|(0<<WDP3)|(1<<WDP2)|(0<<WDP1)|(1<<WDP0)) ; 0,5 s timeout
                ldi     temp1,((0<<WDCE)|(1<<WDE)|(0<<WDP3)|(1<<WDP2)|(1<<WDP1)|(1<<WDP0)) ; 2,0 s timeout
                sts     WDTCSR,temp1                    ; turn on WDT within four instruction cycles
                sei                                     ; global interrupts back on
                ret

                in r0, _SFR_IO_ADDR(SREG)               ; save SREG
                push r0                                 ; save SREG on stack
                cli                                     ; disable interrupts

                pop r0                                  ; restore previous SREG
                out _SFR_IO_ADDR(SREG), r0              ; restore SREG and interrupt status
                ret

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void issue_wdt_reset(void)                                                                                      |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Issue watchdog timer reset                                                                                            |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
.global issue_wdt_reset
issue_wdt_reset:
                cli                                     ; turn off global interrupts
                ldi     temp1,((1<<WDCE)|(1<<WDE))      ; set WDCE and WDE
                sts     WDTCSR,temp1
                ldi     temp1,((0<<WDCE)|(1<<WDE)|(0<<WDP3)|(0<<WDP2)|(0<<WDP1)|(0<<WDP0)) ; 16 ms timeout
                sts     WDTCSR,temp1                    ; turn on WDT within four instruction cycles
loop:           rjmp    loop                            ; WDT reset is issued after 512 cycles (16 ms)

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void set_system_clock_prescaler(void)                                                                           |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Set system clock prescaler                                                                                            |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
.global set_system_clock_prescaler
set_system_clock_prescaler:
                ldi     temp1,(1<<CLKPCE)
                sts     CLKPR,temp1                     ; set Clock Prescaler Change Enable
                ldi     temp1,(0<<CLKPS3)|(0<<CLKPS2)|(0<<CLKPS1)|(0<<CLKPS0) ; divide clock by 1
                sts     CLKPR,temp1                     ; set prescaler
                ret

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void iv2app(void)                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Move interrupt vectors to application space                                                                           |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
.global iv2app
iv2app:         in r0, _SFR_IO_ADDR(SREG)               ; save SREG
                push r0                                 ; save SREG on stack
                cli                                     ; disable interrupts

                in temp1, _SFR_IO_ADDR(MCUCR)           ; Get MCUCR
                mov temp2, temp1
                ori temp1, (1<<IVCE)                    ; Enable change of Interrupt Vectors
                out _SFR_IO_ADDR(MCUCR), temp1          ; Move interrupts to Boot Flash section
                andi temp2, ~(1<<IVSEL)
                out _SFR_IO_ADDR(MCUCR), temp2

                pop r0                                  ; restore previous SREG
                out _SFR_IO_ADDR(SREG), r0              ; restore SREG and interrupt status
                ret

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void iv2boot(void)                                                                                              |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Move interrupt vectors to boot space                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
.global iv2boot
iv2boot:        in r0, _SFR_IO_ADDR(SREG)               ; save SREG
                push r0                                 ; save SREG on stack
                cli                                     ; disable interrupts

                in temp1, _SFR_IO_ADDR(MCUCR)           ; Get MCUCR
                mov temp2, temp1
                ori temp1, (1<<IVCE)                    ; Enable change of Interrupt Vectors
                out _SFR_IO_ADDR(MCUCR), temp1          ; Move interrupts to Boot Flash section
                ori temp2, (1<<IVSEL)
                out _SFR_IO_ADDR(MCUCR), temp2

                pop r0                                  ; restore previous SREG
                out _SFR_IO_ADDR(SREG), r0              ; restore SREG and interrupt status
                ret

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       bool flash_erase(uint16_t page_number)                                                                          |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Erase flash memory page. Code from ATMEGA644 data sheet.                                                              |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
.global flash_erase
flash_erase:    in r0, _SFR_IO_ADDR(SREG)               ; save SREG
                push r0                                 ; save SREG on stack
                cli                                     ; disable interrupts

                mov zh, r24     ;move page number to zh
                ldi zl,0        ;clear zl
                ldi spmcrval, (1<<PGERS) | (1<<SPMEN)  ;Store Program Memory Enable and Page Erase
                call Do_spm
                ldi spmcrval, (1<<RWWSRE) | (1<<SPMEN) ;re-enable the RWW section
                call Do_spm
                ldi r24,1       ;return true

                pop r0                                  ; restore previous SREG
                out _SFR_IO_ADDR(SREG), r0              ; restore SREG and interrupt status
                ret

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       bool flash_program(uint16_t page_number, uint8_t *page_buffer)                                                  |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Program flash memory page. Code from ATMEGA644 data sheet.                                                            |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
.global flash_program
flash_program:  in r0, _SFR_IO_ADDR(SREG)               ; save SREG
                push r0                                 ; save SREG on stack
                cli                                     ; disable interrupts

                push r1
                push r28
                push r29
                mov zh, r24     ;move page number to zh
                ldi zl,0        ;clear zl
                mov yl,r22      ;pointer to page buffer (low)
                mov yh,r23      ;pointer to page buffer (high)

; transfer data from RAM to Flash page buffer
                ldi looplo, lo8(PAGESIZEB) ;init loop variable
                ldi loophi, hi8(PAGESIZEB) ;not required for PAGESIZEB<=256
Wrloop:
                ld r0,y+
                ld r1,y+
                ldi spmcrval, (1<<SPMEN)
                call Do_spm     ;store r1:r0 in page buffer addressed by Z-pointer
                adiw zl,2
                sbiw looplo,2   ;use subi for PAGESIZEB<=256
                brne Wrloop
; execute Page Write
                subi zl, lo8(PAGESIZEB)         ;restore pointer
                sbci zh, hi8(PAGESIZEB)         ;not required for PAGESIZEB<=256
                ldi spmcrval, (1<<PGWRT) | (1<<SPMEN)
                call Do_spm     ;page write
; re-enable the RWW section
                ldi spmcrval, (1<<RWWSRE) | (1<<SPMEN)
                call Do_spm     ;will wait for completion of previous command
; read back and check, optional
                ldi looplo, lo8(PAGESIZEB)      ;init loop variable
                ldi loophi, hi8(PAGESIZEB)      ;not required for PAGESIZEB<=256
                subi YL, lo8(PAGESIZEB)         ;restore pointer
                sbci YH, hi8(PAGESIZEB)
Rdloop:
                lpm r0, Z+
                ld r1, Y+
                cpse r0, r1
                rjmp Error
                sbiw looplo, 1                  ;use subi for PAGESIZEB<=256
                brne Rdloop
                ldi r24,1       ;return true (no error)
; return to RWW section
; verify that RWW section is safe to read
Return:
                in temp1, _SFR_IO_ADDR(SPMCSR)
                sbrs temp1, RWWSB ; If RWWSB is set, the RWW section is not ready yet
                rjmp EiRet

; re-enable the RWW section
                ldi spmcrval, (1<<RWWSRE) | (1<<SPMEN)
                call Do_spm
                rjmp Return

Error:          ldi r24,0       ;return false (error)
                ldi r25,0       ;return error
                rjmp Return

EiRet:          pop r29
                pop r28
                pop r1
                pop r0                                  ; restore previous SREG
                out _SFR_IO_ADDR(SREG), r0              ; restore SREG and interrupt status
                ret

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Do_spm                                                                                                          |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Used by flash_erase and flash_program. Code from ATMEGA644 data sheet.                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
Do_spm:
; check for previous SPM complete
Wait_spm:       in temp1, _SFR_IO_ADDR(SPMCSR)
                sbrc temp1, SPMEN
                rjmp Wait_spm

; input: spmcrval determines SPM action
; disable interrupts if enabled, store status
                in temp2, _SFR_IO_ADDR(SREG)
                cli
; check that no EEPROM write access is present
Wait_ee:        sbic _SFR_IO_ADDR(EECR), EEPE
                rjmp Wait_ee

; SPM timed sequence
                out _SFR_IO_ADDR(SPMCSR), spmcrval
                spm
; restore SREG (to enable interrupts if originally enabled)
                out _SFR_IO_ADDR(SREG), temp2
                ret

;.global change_clock
;
;change_clock:
;                LDI     R25,0x80        ;
;                STS     CLKPR,R25       ;enable updating of CLKPR
;                STS     CLKPR,R24       ;write the new value
;                ret                     ;
;
;.global add_two
;
;add_two:
;                ADD R24,R22             ;val_1 = val_1 + val_2
;                CLR     R25             ;return val_1
;                ret                     ;
;
;.global TIMER0_OVF_vect
;
;TIMER0_OVF_vect:
;                sbi _SFR_IO_ADDR(PIND),0        ;toggle PD.0
;                reti                            ;and done

