/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|          Program Name: Antennenumschalter                                                                             |
|                Author: Michael Krmer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 26-Nov-2020                                                                                    |
|                  Tool: AVR Studio 7.0                                                                                 |
|                Device: ATMega644PA-AU                                                                                 |
|               Version: 0.1                                                                                            |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Antenna switch                                                                                              |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+

Preprocessor-Macros. Define the following preprocessor symbols to achieve the respective behavior:
====================

SIMULATION:     If SIMULATION is defined, certain hardware waits and optimizations are turned off.
                SIMULATION must be undefined for final hardware.

DEBUGCODE:      If DEBUGCODE is defined, some code segments for debugging will be enabled.
                DEBUGCODE should be undefined for the release version.

USE_TEMPSENSOR: Use TMP275 temperature sensor
USE_ADC:        Use AD-converter
BOARD:          Board number as defined in ModbusBoards.h
                See the project specific include file for these settings

Fuses:          For released board (set as required for development phase)
======          Brown-out detection at VCC=4.3 V (for VCC=5.0V)
                Brown-out detection at VCC=2.7 V (for VCC=3.3V)
                OCDEN/JTAGEN disabled (unchecked)
                SPIEN enabled (checked)
                WDTON disabled (unchecked)
                EESAVE enabled (checked)
                BOOTSZ "Boot Flash size=4096 words, Boot address=$7000"
                BOOTRST disabled (unchecked)
                CKDIV8 disabled (unchecked, note that CKDIV8 prescaler ist setup properly during initialization)
                CKOUT disabled (unchecked)
                SUT_CKSEL "Full Swing Oscillator; Start-up time: 16K CK + 65 ms; Crystal Osc.; slowly rising power"

                Fuse Reg:    EXTENDED: 0xFC (for 3.3V operation: 0xFD)
                =========    HIGH:     0xD1
                             LOW:      0xF7

Notes:          1.) Nested interrupts are not permitted! It is assumed that an interrupt handler is never interrupted.
======          2.) Never ship a product with the OCDEN Fuse programmed regardless of the setting of Lock bits and JTAGEN
                    Fuse. A programmed OCDEN Fuse enables some parts of the clock system to be running in all sleep modes.
                    This may increase the power consumption.

ToDo:
=====

*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define DECLARE_VARS /* This is the main program in which the global variables are declared */
#include "project.h"

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define BLSW_VERSION 1000 /* Software version of bootloader (16 bit). Displayed as nn.mmm */

#ifdef TestBench
#define FC_RCO 0x01 /* Function code: Read Coils */
#define FC_RDI 0x02 /* Function code: Read Discrete Inputs */
#define FC_RHR 0x03 /* Function code: Read Holding Register */
#define FC_RIR 0x04 /* Function code: Read Input Register */
#define FC_WSC 0x05 /* Function code: Write Single Coil */
#define FC_WSR 0x06 /* Function code: Write Single Register */
#define FC_WMR 0x10 /* Function code: Write Multiple Register */
#define MB_ERROR 0x80 /* Modbus response error flag */
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool page_erase(uint16_t addr, uint16_t *data) ;
bool page_program(uint16_t addr, uint16_t *data) ;
bool flash_erase(uint16_t page_number) ;
bool flash_program(uint16_t page_number, uint8_t *page_buffer) ;
bool check_flash(void) ;

#ifdef TestBench
void TB_RX_Message(uint8_t device, uint8_t func, uint16_t addr, char *msg, uint8_t len, uint16_t crc) ;
bool TB_TX_Message(uint8_t device, uint8_t func, char *msg, uint8_t len, uint16_t crc) ;
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       EEPROM data                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       PROGMEM constants                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

                              // ,--------------------------------------------------------------------------------------------------------- modbus start address
                              // |                          ,------------------------------------------------------------------------------ number of elements
                              // |                          |           ,------------------------------------------------------------------ type of register
                              // |                          |           |                     ,-------------------------------------------- direct value, pointer to value or physical address in case of EEPROM or PROGMEM
                              // |                          |           |                     |                                ,----------- pointer to read function
const PROGMEM MODBUS_REGISTER // |                          |           |                     |                                |     ,----- pointer to write function
      Input_Register[] = {    // |                          |           |                     |                                |     |
                                {MA_IR + IRO_XTAL_FRQ,      1,          T_IMMEDIATE | T_WORD, CPU_INCLK_KHZ,                   NULL, NULL},  /* crystal frequency in kHz */
                                {MA_IR + IRO_MODULE_TYPE,   1,          T_IMMEDIATE | T_WORD, MODULE_TYPE | BOOTLOADER_TYPE,   NULL, NULL},  /* Module type */
                                {MA_IR + IRO_MODULE_NUMBER, 1,          T_IMMEDIATE | T_WORD, BOARD,                           NULL, NULL},  /* Module number */
                                {MA_IR + IRO_HW_VERSION,    1,          T_IMMEDIATE | T_WORD, HW_VERSION,                      NULL, NULL},  /* Hardware version (16 bit). Displayed as nn.mmm */
                                {MA_IR + IRO_SW_VERSION,    1,          T_IMMEDIATE | T_WORD, BLSW_VERSION,                    NULL, NULL},  /* Software version (16 bit). Displayed as nn.mmm */
                                {MA_IR + IRO_CPU_CLK,       1,          T_IMMEDIATE | T_WORD, CPU_CLK,                         NULL, NULL},  /* CPU clock speed in kHz */
                                {MA_IR + IRO_IO_CLK,        1,          T_IMMEDIATE | T_WORD, IO_CLK,                          NULL, NULL},  /* I/O clock speed in kHz */
                                {MA_IR + IRO_APP_PAGECNT,   1,          T_IMMEDIATE | T_WORD, APP_PAGECNT,                     NULL, NULL},  /* Number of pages for application code */
                                {MA_IR + IRO_BOOT_PAGECNT,  1,          T_IMMEDIATE | T_WORD, BOOT_PAGECNT,                    NULL, NULL},  /* Number of pages for bootloader code */
                                {MA_EEP,                    EEP_SIZEW,  T_EEPPTR    | T_WORD, 0,                               NULL, NULL},  /* EEPROM is mapped here */
                                {MA_FLASH,                  FLASH_SIZEW,T_PROGPTR   | T_WORD, 0,                               NULL, NULL},  /* Flash memory is mapped here */
                                } ;

                              // ,----------------------------------------------------------------------------------------------------------------------- modbus start address
                              // |                             ,----------------------------------------------------------------------------------------- number of elements
                              // |                             |            ,---------------------------------------------------------------------------- type of register
                              // |                             |            |                         ,-------------------------------------------------- direct value, pointer to value or physical address in case of EEPROM or PROGMEM
                              // |                             |            |                         |                                  ,--------------- pointer to read function
const PROGMEM MODBUS_REGISTER // |                             |            |                         |                                  |         ,----- pointer to write function
      Holding_Register[] = {  // |                             |            |                         |                                  |         |
                                {MA_SYSREG + SREGO_RESTART,    1,           T_EXECUTE,                0,                                 NULL,     start_application}, /* Write MAGIC_NUMBER to start application */
                                {MA_SYSREG + SREGO_STARTBOOT,  1,           T_EXECUTE,                0,                                 NULL,     start_bootloader},  /* Restart bootloader */
                                {MA_SYSREG + SREGO_PROGENABLE, 1,           T_EXECUTE,                0,                                 read_to,  prog_enable},       /* Read timeout / Programming enable */
                                {MA_SYSREG + SREGO_NOAPPSTART, 1,           T_DATAPTR,                (uint16_t) &noapp_start,           NULL,     NULL},              /* Don't start application from bootloader if value is not zero */
//                               |                             |                                      |                                  |         |
                                {MA_BOOT + BOOTO_ERASE,        1,           T_EXECUTE,                0,                                 NULL,     page_erase},        /* Erase page */
                                {MA_BOOT + BOOTO_PROG,         1,           T_EXECUTE,                0,                                 NULL,     page_program},      /* Program page */
                                {MA_BOOTPB + BOOTPBO_PB,       PAGESIZEW,   T_DATAPTR | T_WORD,       (uint16_t) &page_buffer[0],        NULL,     NULL},              /* Read or write page buffer */
//                               |                             |                                      |                                  |         |
                                {MA_EEPROM,                    EEP_SIZEW, T_EEPPTR | T_WORD,          0,                                 NULL,     NULL},              /* Read or write EEPROM (in words) */
        } ;

const PROGMEM MODBUS_DESCRIPTOR modbus_desc = {
        (PROGMEM MODBUS_REGISTER *) Input_Register,   sizeof(Input_Register)/sizeof(Input_Register[0]),
        (PROGMEM MODBUS_REGISTER *) Holding_Register, sizeof(Holding_Register)/sizeof(Holding_Register[0]),
        NULL, 0,
        NULL, 0
        } ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Inline functions                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/
__inline static void eeprom_write_byte_cond(uint8_t  *addr, uint8_t  data) {if (eeprom_read_byte(addr) != data) eeprom_write_byte(addr, data) ; }
__inline static void eeprom_write_word_cond(uint16_t *addr, uint16_t data) {if (eeprom_read_word(addr) != data) eeprom_write_word(addr, data) ; }

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Main                                                                                                            |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| MAIN program.                                                                                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
int main(void)
{ /* executed on coldstart */
#ifdef TestBench
bool rc ;
#endif

   initial_things() ; /* Do initialization after a reset occured */
   init_CPU() ; /* initialize CPU, turn global interrupts on */

   if ((reset_flags & (1<<PORF)) != 0)
   { /* This is a cold start, check application flash */
      application_is_healthy = check_flash() ;
   }

/* application is corrupt or no coldstart (i.e. bootloader was called from application) */

   wdt_on() ; /* switch watchdog timer on */

#ifdef SIMULATION
///* some test code, may be removed later */
//   init_CPU() ; /* initialize CPU, turn global interrupts on */
//   /* Initialise EEPROM with default values, for debugging */
//   mcb_eepinit(&eemcb[0], &default_mcb) ;

#ifdef TestBench
   rc = false ;

// Write multiple registers, 4 page buffer registers
   TB_RX_Message(MODBUS_DEVICEADDR, FC_WMR, MA_BOOTPB + BOOTPBO_PB + 32, "\x00\x04\x08\x94\x0c\x03\x44\x94\x0c\x03\x56", 11, 0x3d8d) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_WMR, "\x17\xd4\x00\x04", 4, 0x8683) ;

// Write single register, flash page lschen
   TB_RX_Message(MODBUS_DEVICEADDR, FC_WSR, MA_BOOT + BOOTO_ERASE, "\x39\xf1", 2, 0xfd21) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_WSR, "\x17\x72\x39\xf1", 4, 0xfd21) ;

// Write single register, flash page schreiben
   TB_RX_Message(MODBUS_DEVICEADDR, FC_WSR, MA_BOOT + BOOTO_PROG, "\x00\x01", 2, 0x1f60) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_WSR, "\x17\x71\x39\xf1", 4, 0x0d21) ;

// Read number of resets
   TB_RX_Message(MODBUS_DEVICEADDR, FC_RHR, MA_PARMS + STRO_RESET, "\x00\x01", 2, 0x15cc) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_RHR, "\x02\x00\x01", 3, 0xb847) ;

// Write multiple registers, 8 RTC registers
   TB_RX_Message(MODBUS_DEVICEADDR, FC_WMR, MA_RTC + RTCO_YEAR, "\x00\x08\x10\x00\x16\x00\x03\x00\x07\x00\x10\x00\x34\x00\x2d\x00\x46\x00\x00", 19, 0x7a20) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_WMR, "\x0b\xc2\x00\x08", 4, 0x6087) ;

// Read holding registers, 8 RTC registers
   TB_RX_Message(MODBUS_DEVICEADDR, FC_RHR, MA_RTC + RTCO_YEAR, "\x00\x08", 2, 0xe544) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_RHR, "\x10\x00\x16\x00\x03\x00\x07\x00\x10\x00\x34\x00\x2d\x00\x46\x00\x00", 17, 0x5cf3) ;

// Read input register
   TB_RX_Message(MODBUS_DEVICEADDR, FC_RIR, MA_IR + IRO_XTAL_FRQ, "\x00\x01", 2, 0x3328) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_RIR, "\x02\x27\x10", 3, 0x36ce) ;

// Read 2 RTC registers
   TB_RX_Message(MODBUS_DEVICEADDR, FC_RHR, MA_RTC + RTCO_YEAR, "\x00\x02", 2, 0x65bc) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_RHR, "\x04\x00\x16\x00\x03", 5, 0xb537) ;

// Read MCB Register beyond end (should fail)
   TB_RX_Message(MODBUS_DEVICEADDR, FC_RHR, MA_MCB+ MCBO_PGRCNTL, "\x00\x02", 2, 0x05a4) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, MB_ERROR|FC_RHR, "\x02", 1, 0xf137) ;

// Read 5 MCB registers
   TB_RX_Message(MODBUS_DEVICEADDR, FC_RHR, MA_MCB + MCBO_MODBUSDEVID, "\x00\x05", 2, 0x05a7) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_RHR, "\x0a\x00\x1e\x03\xc0\x00\x00\x00\x00\x00\x01", 11, 0xce50) ;

// Read 1 baudrate register
   TB_RX_Message(MODBUS_DEVICEADDR, FC_RIR, MA_BAUD + 1, "\x00\x01", 2, 0x42ea) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_RIR, "\x02\x00\x1e", 3, 0xacfa) ;

// Write single register, MCB baudrate
   TB_RX_Message(MODBUS_DEVICEADDR, FC_WSR, MA_MCB + MCBO_BAUDRATE, "\x07\x80", 2, 0x5bf4) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_WSR, "\x0b\xb9\x07\x80", 4, 0x5bf4) ;

// Write single register, invalid MCB baudrate
   TB_RX_Message(MODBUS_DEVICEADDR, FC_WSR, MA_MCB + MCBO_BAUDRATE, "\x07\x60", 2, 0x5a7c) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, MB_ERROR|FC_WSR, "\x02", 1, 0xf267) ;

// Read TCXO registers
   TB_RX_Message(MODBUS_DEVICEADDR, FC_RHR, MA_PARMS + 0, "\x00\x01", 2, 0x44e8) ;
   rc |= TB_TX_Message(MODBUS_DEVICEADDR, FC_RHR, "\x02\xff\xff", 3, 0x2c36) ;

#endif
#endif

   listen_only = false ;
   restart_communications = false ;

   for (;;)
   {
/* Setup serial receiver and wait for message */
      receive(&rxbuf) ;
      wait_for_rx(&rxbuf) ; /* wait for received message */

      switch(rxbuf.status)
      { /* reception completed. Check for error, otherwise check address and prepare response */
         case RXSTATE_COMPLETED:
            modbus_message_received(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;
         break ;

         case RXSTATE_ERROR:
         case RXSTATE_BUFFEROVERFLOW :
         /* error, just increment error counter */
            com_error_count_l++ ;
            if (com_error_count_l == 0) com_error_count_h++ ;
         break ;

         default: /* ignore other cases for now */
         break ;
      }
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void wait_for_rx_user(void)                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| User specific wait tasks.                                                                                             |
|                                                                                                                       |
| This function executes regularly while waiting for a bus transaction. Interrupts shall not be disabled.               |
| Bus messages may be lost, if execution takes too long.                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void wait_for_rx_user(void)
{
   if ((!startup_delay.active) && (!noapp_start) && application_is_healthy)
   { /* startup timer is expired and application start is not inhibited and application is healthy  */
      __asm("cli") ; /* Turn off global interrupt */
      iv2app() ;     /* move interrupt vectors to application space */
      goto *NULL ;   /* Jump to application Reset vector */
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       page_erase                                                                                                      |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Erase specific flash memory page or all pages if page == 0xffff                                                       |
|                                                                                                                       |
| Returns true if no error occured, false otherwise.                                                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool page_erase(uint16_t addr, uint16_t *page)
{
uint16_t i ;
   if (prog_timer.active)
   {
      start_timer(&prog_timer, PROG_TIMEOUT) ;
      if (*page == 0xffff)
      { /* erase all pages */
         for (i=0; i<APP_PAGECNT; i++) if (!flash_erase(i)) return(false) ;
         return(true) ;
      }
      else
      if (*page >= APP_PAGECNT) return(false) ;
      return(flash_erase(*page)) ;
   }
   return(false) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       page_program                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Program flash memory page with data from "page_buffer"                                                                |
|                                                                                                                       |
| Returns true if no error occured, false otherwise.                                                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool page_program(uint16_t addr, uint16_t *page)
{
   if (prog_timer.active)
   {
      start_timer(&prog_timer, PROG_TIMEOUT) ;
      if (*page >= APP_PAGECNT) return(false) ;
      return(flash_program(*page, page_buffer)) ;
   }
   return(false) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       check_flash                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Check sanity of application flash memory                                                                              |
|                                                                                                                       |
| Returns true if healthy.                                                                                              |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool check_flash(void)
{
uint16_t start, size ;
SIGNATURE *sig ;
   sig = (SIGNATURE *) (SIGNATUREPAGE * PAGESIZEB) ; /* byte address for subsequent signature accesses */
   size = sizeof(*sig) - sizeof(sig->sig_crc) ;
   if (crc16_progmem((uint8_t *) sig, size) == pgm_read_word(&sig->sig_crc))
   { /* signature is valid */
      start = pgm_read_word(&sig->app_start) ;
      size = pgm_read_word(&sig->app_end) - pgm_read_word(&sig->app_start) + 1 ;
      if (crc16_progmem((uint8_t *) start, size) == pgm_read_word(&sig->app_crc))
      { /* CRC of application confirmed */
         return(true) ;
      }
   }
   return(false) ;
}

#ifdef TestBench
/*
+-----------------------------------------------------------------------------------------------------------------------+
|       TB_RX_Message                                                                                                   |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Receive and process ModBus message                                                                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void TB_RX_Message(uint8_t device, uint8_t func, uint16_t addr, char *msg, uint8_t len, uint16_t crc)
{
uint8_t i ;
   rxbuf.buffer[0]  = device ;  /* slave address */
   rxbuf.buffer[1]  = func ;    /* function code */
   rxbuf.buffer[2]  = addr>>8 ; /* address high */
   rxbuf.buffer[3]  = addr&0xff ; /* address low */
   for (i=0; i<len; i++)
   {
      rxbuf.buffer[i+4]  = msg[i] ; /* append message */
   }

   rxbuf.buffer[i+4]  = (crc>>8) ; /* crc high */
   rxbuf.buffer[i+5]  = (crc&0xff) ; /* crc low */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       TB_TX_Message                                                                                                   |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Analyze transmit message. Return true if on error.                                                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool TB_TX_Message(uint8_t device, uint8_t func, char *msg, uint8_t len, uint16_t crc)
{
uint8_t i ;
bool diff ;
   diff = false ;
   diff = diff || (txbuf.buffer[0] != device) ;  /* slave address */
   diff = diff || (txbuf.buffer[1] != func) ;    /* function code */
   for (i=0; i<len; i++)
   {
      diff = diff || (txbuf.buffer[i+2] != msg[i]) ; /* check message */
   }

   diff = diff || (txbuf.buffer[i+2] != (crc>>8)) ; /* crc high */
   diff = diff || (txbuf.buffer[i+3] != (crc&0xff)) ; /* crc low */
   return(diff) ;
}

#endif


