﻿/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|             File Name: ModbusCommon.h                                                                                 |
|                Author: Michael Krämer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 26.02.2022                                                                                     |
|                  Tool: AVR Studio 7.0 and Visual Studio                                                               |
|                Device: ATTiny, ATMEGA, ATXMEGA                                                                        |
|               Version: 0.1                                                                                            |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Common include file for modbus software (used for host and remote software)                                 |
|           Defines common Modbus addresses and data structures                                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#ifndef MODBUSCOMMON

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define MODBUSCOMMON 1
#define MAGIC_NUMBER 0x39f1     /* magic number to restart remote device */

#define MODBUSADDR_PORTS 1000   /* modbus address of ports for Modbus addressing */
#define MODBUSADDR_IR  2000     /* modbus address of input registers for Modbus addressing (see input_registers in PROGMEM) */
#define MODBUSADDR_HR 3000      /* modbus address of holding (input/output) registers for Modbus addressing (see below) */
#define MODBUSADDR_FR 1000      /* File Records */

#define MA_PORTS  (MODBUSADDR_PORTS +    0) /* Modbus address for I/O port access ("coils") */
#define MA_IR     (MODBUSADDR_IR    +    0) /* Modbus address of input registers */
#define MA_BAUD   (MODBUSADDR_IR    +   10) /* Modbus address to read supported baudrates */
#define MA_EEP    (MODBUSADDR_IR    +  100) /* Modbus address to read EEPROM contents */
#define MA_PRIVIR (MODBUSADDR_IR    +  200) /* Modbus address of private input registers */
#define MA_FLASH  (MODBUSADDR_IR    + 2000) /* Modbus address to read FLASH memory contents */
#define MA_MCB    (MODBUSADDR_HR    +    0) /* Modbus address of MCB entry (which is a copy from the current EEPROM MCB) */
#define MA_RTC    (MODBUSADDR_HR    +   10) /* Modbus address of first RTC register */
#define MA_SYSREG (MODBUSADDR_HR    +   20) /* Registers for system functions, e.g. restart device */
#define MA_STATUS (MODBUSADDR_HR    +  100) /* Start address for virtual status registers */
#define MA_TELEM  (MODBUSADDR_HR    +  200) /* Start address for telemetric data, temperature, voltage etc. */
#define MA_FUNC   (MODBUSADDR_HR    +  300) /* Start address for function registers */
#define MA_PARMS  (MODBUSADDR_HR    +  500) /* Start address for configurable EEPROM parameter registers */
#define MA_PRIVHR (MODBUSADDR_HR    +  800) /* Start address for private holding registers */
#define MA_TEST   (MODBUSADDR_HR    + 1000) /* Start address for test data */
#define MA_BOOT   (MODBUSADDR_HR    + 3000) /* Start address for bootloader commands */
#define MA_BOOTPB (MODBUSADDR_HR    + 3100) /* Start address for bootloader page buffer */
#define MA_EEPROM (MODBUSADDR_HR    + 4000) /* Start address for EEPROM programming */
#define MA_FREC   (MODBUSADDR_FR    +    0) /* Start address for file records */

/* Input registers starting at MA_IR */
enum {IRO_XTAL_FRQ=0,   /* Input register offset to crystal frequency in kHz */
      IRO_MODULE_TYPE,  /* Input register offset to Module type */
      IRO_MODULE_NUMBER,/* Input register offset to Module number */
      IRO_HW_VERSION,   /* Input register offset to Hardware version (16 bit). Displayed as nn.mmm */
      IRO_SW_VERSION,   /* Input register offset to Software version (16 bit). Displayed as nn.mmm */
      IRO_CPU_CLK,      /* Input register offset to CPU clock speed in kHz */
      IRO_IO_CLK,       /* Input register offset to I/O clock speed in kHz */
      IRO_APP_PAGECNT,  /* Number of pages in RWW section of this device, used for application code */
      IRO_BOOT_PAGECNT, /* Number of pages in RWW section of this device, used for bootloader code */
      IR_COUNT} ;       /* Number of input registers */

/* Common EEPROM Contents (MCB) starting at MA_MCB */
enum {MCBO_NUMREG=0,    /* Number of MCB registers */
      MCBO_MODBUSDEVID, /* Offset to ModBus device ID in MCB */
      MCBO_BAUDRATE,    /* Offset to baudrate in MCB */
      MCBO_RTCCORR,     /* Offset to RTC correction in MCB */
      MCBO_PGRCNTH,     /* Offset to programming counter (high) in MCB */
      MCBO_PGRCNTL,     /* Offset to programming counter (low) in MCB */
      MCBC_COUNT} ;     /* Number of common MCB registers */

/* Holding Registers (RTC) starting at MA_RTC */
enum {RTCO_YEAR=0,      /* Offset to RTC year */
      RTCO_MONTH,       /* Offset to RTC month */
      RTCO_DAY,         /* Offset to RTC day */
      RTCO_HOUR,        /* Offset to RTC hour */
      RTCO_MIN,         /* Offset to RTC minute */
      RTCO_SEC,         /* Offset to RTC seconds */
      RTCO_CS,          /* Offset to RTC centiseconds */
      RTCO_ENABLE,      /* Offset to Enable RTC */
      RTC_COUNT} ;      /* Number of RTC registers */

/* Holding Registers (System Registers) starting at MA_SYSREG */
enum {SREGO_RESTART=0,  /* Write MAGIC_NUMBER to restart device */
      SREGO_CLR_CNT,    /* Write MAGIC_NUMBER to clear error counters */
      SREGO_RESDEVID,   /* Reset device ID in EEPROM to recover from multiple devices with same ID */
      SREGO_STARTBOOT,  /* Start bootloader */
      SREGO_PROGENABLE, /* Enable Flash programming */
      SREGO_NOAPPSTART, /* Don't start application from bootloader if value is not zero */
      SYSREG_COUNT} ;   /* Number of system registers */

/* Holding Registers (Bootloader Registers) starting at MA_BOOT */
enum {BOOTO_ERASE=0,    /* Write page number to erase page */
      BOOTO_PROG,       /* Write page number to program page with data from page buffer */
      BOOTO_COUNT} ;    /* Number of bootloader registers */

/* Holding Registers (Bootloader Page Buffer Registers) starting at MA_BOOTPB */
enum {BOOTPBO_PB=0,     /* Read/Write data to page buffer */
      BOOTPBO_COUNT} ;  /* Number of page buffer registers */

/* Holding Registers (Status Registers) starting at MA_STATUS */
enum {STRO_MCUSR=0,     /* Offset to MCU Status Register */
      STRO_RESET,       /* Offset to total number of resets since last power-on reset */
      STRO_WDTR,        /* Offset to number of watchdog timer resets since last power-on reset */
      STRO_SR,          /* Offset to number of software resets since last power-on reset */
      STRO_EXTR,        /* Offset to number of external resets since last power-on reset */
      STRO_BOR,         /* Offset to number of brown-out resets since last power-on reset */
      STRO_TMSGH,       /* Offset to total number of error free messages since last power-on reset (higher) */
      STRO_TMSGL,       /* Offset to total number of error free messages since last power-on reset (lower) */
      STRO_MSGH,        /* Offset to number of messages for this module since last power-on reset (higher) */
      STRO_MSGL,        /* Offset to number of messages for this module since last power-on reset (lower) */
      STRO_CRCEH,       /* Offset to number of CRC errors since last power-on reset (higher) */
      STRO_CRCEL,       /* Offset to number of CRC errors since last power-on reset (lower) */
      STRO_COMH,        /* Offset to number of communication errors since last power-on reset (higher) */
      STRO_COML,        /* Offset to number of communication errors since last power-on reset (lower) */
      STRO_RXCNTH,      /* Offset to number of received bytes since last power-on reset (higher) */
      STRO_RXCNTL,      /* Offset to number of received bytes since last power-on reset (lower) */
      STRO_TXCNTH,      /* Offset to number of transmitted bytes since last power-on reset (higher) */
      STRO_TXCNTL,      /* Offset to number of transmitted bytes since last power-on reset (lower) */
      STRO_ONTIMEH,     /* Offset to power-on time in centi-seconds (higher) */
      STRO_ONTIMEM,     /* Offset to power-on time in centi-seconds (middle) */
      STRO_ONTIMEL,     /* Offset to power-on time in centi-seconds (lower) */
      STR_COUNT} ;      /* Number of status registers */

/* Holding Registers (Telemetrie Registers) stating at MA_TELEM */
enum {TREGO_TEMP=0,     /* Offset to Temperature Register */
      TREGO_VSUPP,      /* Offset to supply voltage Register */
      TREG_COUNT} ;     /* Number of telemetrie registers */

/* File Records starting at MA_FREC */
enum {FRECO_MAINFN=0,   /* File name of main source file */
      FRECO_MAINTS,     /* Time stamp of main source file (Sun Sep 16 01:03:52 1973) */
      FRECO_MBFN,       /* File name of main source file */
      FRECO_MBTS,       /* Time stamp of main source file (Sun Sep 16 01:03:52 1973) */
      FRECO_CD,         /* Compile Date (Mmm dd yyyy) */
      FRECO_CT,         /* Compile Time (hh:mm:ss) */
      FREC_COUNT} ;     /* Number of file records */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#ifdef WIN32
/* ensure proper packing under Windows */
//#pragma pack(show)
#pragma pack(push)
#pragma pack(2)
#endif

typedef struct tcxo_struct { /* TCXO parameters */
        uint32_t fn ;   /* parameter fn (nominal crystal frequency in Hz) */
        uint32_t prec ; /* parameter prec (assumed precison of TCXO in ppb) */
        float a0 ;      /* parameter a0 */
        float a1 ;      /* parameter a1 */
        float a2 ;      /* parameter a2 */
        float a3 ;      /* parameter a3 */
        float a4 ;      /* parameter a4 */
        float a5 ;      /* parameter a5 */
        } TCXO ;

typedef struct sig_struct{
        uint16_t app_start ;    /* start address of application code (first byte) */
        uint16_t app_end ;      /* end address of application code (last byte) */
        uint16_t app_crc ;      /* CRC of application code */
        uint16_t sig_crc ;      /* CRC of signature (must be last element in this structure) */
        } SIGNATURE ;

/* Note: AVR supports only float, not double. float is 32-bits. */

#ifdef WIN32
#pragma pack(pop)
//#pragma pack(show)
#endif /* win32 */

typedef struct parm_struct { /* EEPROM parameter block */
        TCXO tcxo ;     /* TCXO parameters */
        bool tcxo_valid ; /* TCXO data read successfully */
        } RPARM ;

#endif
