/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|          Program Name: UART_ATMEGA644.c                                                                               |
|                Author: Michael Krmer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 25-Jan-2021                                                                                    |
|                  Tool: AVR Studio 7.0                                                                                 |
|                Device: ATMEGA644                                                                                      |
|               Version: 0.1                                                                                            |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: UART interface functions for ATMEGA644                                                                      |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+

Preprocessor-Macros. Define the following preprocessor symbols to achieve the respective behaviour:
====================

SIMULATION:  If SIMULATION is defined, certain hardware waits and optimizations are turned off.
             SIMULATION must be undefined for final hardware.

DEBUG:       DEBUG will replace inline functions with standard functions. __inline is not supported
             if optimizations are entirely disabled.

XTAL_FRQ:    Crystal frequency in kHz. Supported values:
             XTAL_FRQ==8000
             XTAL_FRQ==10000
             XTAL_FRQ==11059
             XTAL_FRQ==12000
             XTAL_FRQ==14746
             XTAL_FRQ==18432
             XTAL_FRQ==20000

Note:        Nested interrupts are not permitted!
=====        It is asumed that an interrupt handler is never interrupted.

ToDo:
=====
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#include "project.h"
//#include "modbus.h"
//#include "UART_ATMEGA644.h"

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Inline functions                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       PROGMEM constants                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       init_UART                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Initialize the UART, this function is called at startup.                                                              |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void init_UART(void)
{
/* Setup USART0 for serial data transfer */
/* Set the baud rate */
   UBRR0 = ubrrval ;
//   RX_disable() ; /* disable receiver and RX interrupts, will also flush RX buffer */
//   TX_disable() ; /* disable transmitter and TX interrupts */

/* Disable transmitter and TX interrupts, disable receiver and RX interrupts, will also flush RX buffer */
   UART_disable(RXCIE | TXCIE | UDRIE | RXEN | TXEN | UCSZ2 | TXB8) ;

/* Asynchronous UART, no parity, two stopbits, 8-bit character size, clock polarity unused */
   UCSR0C = ((0<<UMSEL01) | (0<<UMSEL00) | MYUPM01 | (0<<UPM00) | (1<<USBS0) | (1<<UCSZ01) | (1<<UCSZ00) | (0<<UCPOL0)) ;

/* Enable transmitter. Will remain anabled all the time, only transmit interrupts will be toggled. */
   UART_enable(TXEN) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       receive                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Receive serial data to buffer                                                                                         |
|                                                                                                                       |
| Prepare receive buffer and enable UART for reception                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void receive(RXBUFFER *rxb)
{
   prxb = rxb ; /* set global pointer to active receive buffer */
   prxb->next = 0 ;
   prxb->status = RXSTATE_RECEIVING ;
   prxb->size = RXBUFSIZE ;
   prxb->bytecounter = 0 ;
   UART_enable(RXCIE | RXEN) ; /* enable receiver and RX interrupt */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       transmit                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Transmit serial buffer                                                                                                |
|                                                                                                                       |
| Switch on RS485 transmitter and send data from serial buffer. Switch off transmitter when completed                   |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void transmit(TXBUFFER *txb)
{
uint8_t chr ;
   ptxb = txb ; /* set global pointer to active transmit buffer */
   ptxb->next = 0 ;
   if ((!ptxb->completed) && (ptxb->count > 0))
   {
      UART_disable(RXCIE | RXEN) ; /* disable receiver and RX interrupt */
//      UART_enable(TXEN) ; /* enable transmitter */
      RS485_TX_enable() ; /* enable RS485 transmitter (after tx enable to avoid glitch on tx signal) */
      ptxb->bytecounter = 0 ;
      chr = ptxb->buffer[ptxb->next++] ;
      UDR0 = chr ;        /* transmit first character */
      UART_enable(TXCIE | UDRIE) ; /* enable TX interrupts */
   }
   else
   {
      ptxb->completed = true ;
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       USART0_RX                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Interrupt handler for USART0 RX complete                                                                              |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
ISR(USART0_RX_vect)
{
uint8_t chr ;

#ifdef EVEN_PARITY
   if ((UCSR0A & (1<<UPE0)) != 0)
   { /* parity error, just increase error counter */
      com_error_count_l++ ;
      if (com_error_count_l == 0) com_error_count_h++ ;
   }
#endif

   chr = UDR0 ; /* get received data */
   prxb->bytecounter++ ;
   OCR1B = TCNT1+t25 ; /* set character timeout, 1.5 characters plus current character = 2.5 characters = 25 bit times */
   TOINT_enable() ; /* enable timeout interrupts */

   switch(prxb->status)
   {
      case RXSTATE_RECEIVING:
         if (prxb->next < prxb->size)
         {
            prxb->buffer[prxb->next++] = chr ;
         }
         else
         {
            prxb->status = RXSTATE_BUFFEROVERFLOW ;
            TOINT_disable() ; /* disable further timeout interrupts */
            UART_disable(RXCIE | RXEN) ; /* disable receiver and RX interrupt */
         }
      break ;

      case RXSTATE_TIMEOUT:
         prxb->status = RXSTATE_ERROR ; /* we had a character timeout before, but received another character. This is a receive error. */
         TOINT_disable() ; /* disable further timeout interrupts */
         UART_disable(RXCIE | RXEN) ; /* disable receiver and RX interrupt */
      break ;
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       USART0_UDRE                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Interrupt handler for USART0 Data Register empty                                                                      |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
ISR(USART0_UDRE_vect)
{
uint8_t chr ;
   ptxb->bytecounter++ ;
   if ((!ptxb->completed) && (ptxb->next < ptxb->count))
   {
      chr = ptxb->buffer[ptxb->next++] ;
      UDR0 = chr ; /* transmit next character */
   }
   else
   {
      UART_disable(UDRIE) ; /* disable TX DRE interrupt, transmitter and TX complete interrupt remain enabled */
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       USART0_TX                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Interrupt handler for USART0 TX complete                                                                              |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
ISR(USART0_TX_vect)
{
   RS485_TX_disable() ; /* disable RS485 transmitter */
//   UART_disable(TXCIE | UDRIE | TXEN) ; /* disable transmitter and TX interrupts */
   UART_disable(TXCIE | UDRIE) ; /* disable transmitter and TX interrupts */
   ptxb->completed = true ;
   UART_enable(RXCIE | RXEN) ; /* enable receiver and RX interrupt */
}


