/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|          Program Name: Antennentuner                                                                                  |
|                Author: Michael Krmer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 23-Mar-2021                                                                                    |
|                  Tool: AVR Studio 7.0                                                                                 |
|                Device: ATMega644PA-AU                                                                                 |
|               Version: 0.1                                                                                            |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Antenna tuner                                                                                               |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+

Preprocessor-Macros. Define the following preprocessor symbols to achieve the respective behavior:
====================

SIMULATION:     If SIMULATION is defined, certain hardware waits and optimizations are turned off.
                SIMULATION must be undefined for final hardware.

DEBUGCODE:      If DEBUGCODE is defined, some code segments for debugging will be enabled.
                DEBUGCODE should be undefined for the release version.

USE_TEMPSENSOR: Use TMP275 temperature sensor
USE_ADC:        Use AD-converter
BOARD:          Board number as defined in ModbusBoards.h
                See the project specific include file for these settings

Fuses:          For released board (set as required for development phase)
======          Brown-out detection at VCC=4.3 V (for VCC=5.0V)
                Brown-out detection at VCC=2.7 V (for VCC=3.3V)
                OCDEN/JTAGEN disabled (unchecked)
                SPIEN enabled (checked)
                WDTON disabled (unchecked)
                EESAVE enabled (checked)
                BOOTSZ "Boot Flash size=4096 words, Boot address=$7000"
                BOOTRST enabled (checked)
                CKDIV8 disabled (unchecked, note that CKDIV8 prescaler ist setup properly during initialization)
                CKOUT disabled (unchecked)
                SUT_CKSEL "Full Swing Oscillator; Start-up time: 16K CK + 65 ms; Crystal Osc.; slowly rising power"

                Fuse Reg:    EXTENDED: 0xFC (for 3.3V operation: 0xFD)
                =========    HIGH:     0xD0
                             LOW:      0xF7

Notes:          1.) Nested interrupts are not permitted! It is assumed that an interrupt handler is never interrupted.
======          2.) Never ship a product with the OCDEN Fuse programmed regardless of the setting of Lock bits and JTAGEN
                    Fuse. A programmed OCDEN Fuse enables some parts of the clock system to be running in all sleep modes.
                    This may increase the power consumption.

                CKOUT or JTAGEN will make the board unusable for this application, because the respective pins are needed!

ToDo:
=====

*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define DECLARE_VARS /* This is the main program in which the global variables are declared */
#include "project.h"

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool readcoil(uint16_t addr, uint16_t *data) ;
bool writecoil(uint16_t addr, uint16_t *data) ;
bool set_relay(uint16_t addr, uint16_t *data) ;
bool get_relay(uint16_t addr, uint16_t *data) ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       EEPROM data                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/

uint8_t relays_c ; /* current settings of capacitor relays */
uint8_t relays_c_valid = false ; /* validity of relays_c */
uint8_t relays_i ; /* current settings of inductor relays */
uint8_t relays_i_valid = false ; /* validity of relays_i */
uint8_t relays_target ; /* target settings for relays when state machine is started */
uint8_t relay_state = 0 ; /* state counter for relay state machine */
uint8_t next_relay_state = 0 ; /* simple double buffering */
uint8_t next_relay = 0 ; /* sequence number of next relay to be switched */
uint8_t relay_on_highdriver ; /* holds the PORTB bit to switch the high driver on for setting the relay */
uint8_t relay_off_highdriver ; /* holds the PORTB bit to switch the high driver on for resetting the relay */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       PROGMEM constants                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/
char const PROGMEM FileName[] = __FILE__ ;       /* Filename */
char const PROGMEM TimeStamp[] = __TIMESTAMP__ ; /* Sun Sep 16 01:03:52 1973 */
char const PROGMEM CompileDate[] = __DATE__ ;    /* Mmm dd yyyy */
char const PROGMEM CompileTime[] = __TIME__ ;    /* hh:mm:ss */

                              // ,---------------------------------------- modbus start address
                              // |             ,-------------------------- number of elements
                              // |             |   ,---------------------- type of register
                              // |             |   |  ,------------------- direct value, pointer to value or physical address in case of EEPROM or PROGMEM
                              // |             |   |  |  ,---------------- pointer to read function
const PROGMEM MODBUS_REGISTER // |             |   |  |  |          ,----- pointer to write function
      Coils[] =          {    // |             |   |  |  |          |
                                {MA_PORTS + 0, 32, 0, 0, &readcoil, &writecoil},
                                } ;

                              // ,--------------------------------------------------------------------------------------------------------- modbus start address
                              // |                          ,------------------------------------------------------------------------------ number of elements
                              // |                          |           ,------------------------------------------------------------------ type of register
                              // |                          |           |                     ,-------------------------------------------- direct value, pointer to value or physical address in case of EEPROM or PROGMEM
                              // |                          |           |                     |                                ,----------- pointer to read function
const PROGMEM MODBUS_REGISTER // |                          |           |                     |                                |     ,----- pointer to write function
      Input_Register[] = {    // |                          |           |                     |                                |     |
                                {MA_IR + IRO_XTAL_FRQ,      1,          T_IMMEDIATE | T_WORD, CPU_INCLK_KHZ,                   NULL, NULL},  /* crystal frequency in kHz */
                                {MA_IR + IRO_MODULE_TYPE,   1,          T_IMMEDIATE | T_WORD, MODULE_TYPE,                     NULL, NULL},  /* Module type */
                                {MA_IR + IRO_MODULE_NUMBER, 1,          T_IMMEDIATE | T_WORD, BOARD,                           NULL, NULL},  /* Module number */
                                {MA_IR + IRO_HW_VERSION,    1,          T_IMMEDIATE | T_WORD, HW_VERSION,                      NULL, NULL},  /* Hardware version (16 bit). Displayed as nn.mmm */
                                {MA_IR + IRO_SW_VERSION,    1,          T_IMMEDIATE | T_WORD, SW_VERSION,                      NULL, NULL},  /* Software version (16 bit). Displayed as nn.mmm */
                                {MA_IR + IRO_CPU_CLK,       1,          T_IMMEDIATE | T_WORD, CPU_CLK,                         NULL, NULL},  /* CPU clock speed in kHz */
                                {MA_IR + IRO_IO_CLK,        1,          T_IMMEDIATE | T_WORD, IO_CLK,                          NULL, NULL},  /* I/O clock speed in kHz */
                                {MA_IR + IRO_APP_PAGECNT,   1,          T_IMMEDIATE | T_WORD, APP_PAGECNT,                     NULL, NULL},  /* Number of pages for application code */
                                {MA_IR + IRO_BOOT_PAGECNT,  1,          T_IMMEDIATE | T_WORD, BOOT_PAGECNT,                    NULL, NULL},  /* Number of pages for bootloader code */
                                {MA_BAUD,                   CT_SIZE,    T_DATAPTR   | T_WORD, (uint16_t) &supported_baudrates, NULL, NULL},  /* supported baudrates in RAM */
                                {MA_EEP,                    EEP_SIZEW,  T_EEPPTR    | T_WORD, 0,                               NULL, NULL},  /* EEPROM is mapped here */
                                {MA_FLASH,                  FLASH_SIZEW,T_PROGPTR   | T_WORD, 0,                               NULL, NULL},  /* Flash memory is mapped here */
                                } ;

                              // ,------------------------------------------------------------------------------------ file record (file number is always 1)
                              // |                             ,------------------------------------------------------ number of characters or 0 for null-terminated string
                              // |                             |  ,--------------------------------------------------- type of register
                              // |                             |  |          ,---------------------------------------- pointer to file record in PROGMEM
                              // |                             |  |          |                            ,----------- pointer to read function
const PROGMEM MODBUS_REGISTER // |                             |  |          |                            |     ,----- pointer to write function
      File_Records[] = {      // |                             |  |          |                            |     |
                                {MODBUSADDR_FR + FRECO_MAINFN, 0, T_PROGPTR, (uint16_t) FileName,         NULL, NULL},  /* Filename of main source file */
                                {MODBUSADDR_FR + FRECO_MAINTS, 0, T_PROGPTR, (uint16_t) TimeStamp,        NULL, NULL},  /* Time stamp of main source file (Sun Sep 16 01:03:52 1973) */
                                {MODBUSADDR_FR + FRECO_MBFN,   0, T_PROGPTR, (uint16_t) Modbus_FileName,  NULL, NULL},  /* File name of Modbus source file */
                                {MODBUSADDR_FR + FRECO_MBTS,   0, T_PROGPTR, (uint16_t) Modbus_TimeStamp, NULL, NULL},  /* Time stamp of Modbus source file (Sun Sep 16 01:03:52 1973) */
                                {MODBUSADDR_FR + FRECO_CD,     0, T_PROGPTR, (uint16_t) CompileDate,      NULL, NULL},  /* Compile Date (Mmm dd yyyy) */
                                {MODBUSADDR_FR + FRECO_CT,     0, T_PROGPTR, (uint16_t) CompileTime,      NULL, NULL},  /* Compile Time (hh:mm:ss) */
                                } ;

                              // ,------------------------------------------------------------------------------------------------------------ modbus start address
                              // |                            ,------------------------------------------------------------------------------- number of elements
                              // |                            |  ,---------------------------------------------------------------------------- type of register
                              // |                            |  |                         ,-------------------------------------------------- direct value, pointer to value or physical address in case of EEPROM or PROGMEM
                              // |                            |  |                         |                                  ,--------------- pointer to read function
const PROGMEM MODBUS_REGISTER // |                            |  |                         |                                  |         ,----- pointer to write function
      Holding_Register[] = {  // |                            |  |                         |                                  |         |
                                {MA_MCB + MCBO_NUMREG,        1, T_MCBPTR | T_RO,          (uint16_t) &mcb.numreg,            read_mcb, write_mcb},       /* Access number of modbus registers */
                                {MA_MCB + MCBO_MODBUSDEVID,   1, T_MCBPTR,                 (uint16_t) &mcb.module_address,    read_mcb, write_mcb},       /* Access ModBus address in EEPROM */
                                {MA_MCB + MCBO_BAUDRATE,      1, T_MCBPTR | T_WORD,        (uint16_t) &mcb.br10th,            read_mcb, write_mcb},       /* Access baudrate in EEPROM */
                                {MA_MCB + MCBO_RTCCORR,       1, T_MCBPTR | T_WORD,        (uint16_t) &mcb.rtc_correction,    read_mcb, write_mcb},       /* Access RTC correction in EEPROM */
                                {MA_MCB + MCBO_PGRCNTH,       1, T_MCBPTR | T_RO | T_WORD, (uint16_t) &mcb.pgm_cntrh,         read_mcb, write_mcb},       /* Access programming counter (high) in EEPROM */
                                {MA_MCB + MCBO_PGRCNTL,       1, T_MCBPTR | T_RO | T_WORD, (uint16_t) &mcb.pgm_cntrl,         read_mcb, write_mcb},       /* Access programming counter (low) in EEPROM */
//                               |                            |                            |                                  |         |
                                {MA_RTC + RTCO_YEAR,          1, T_DATAPTR,                (uint16_t) &rtc_year,              NULL,     NULL},            /* RTC year */
                                {MA_RTC + RTCO_MONTH,         1, T_DATAPTR,                (uint16_t) &rtc_month,             NULL,     NULL},            /* RTC month */
                                {MA_RTC + RTCO_DAY,           1, T_DATAPTR,                (uint16_t) &rtc_day,               NULL,     NULL},            /* RTC day */
                                {MA_RTC + RTCO_HOUR,          1, T_DATAPTR,                (uint16_t) &rtc_hour,              NULL,     NULL},            /* RTC hour */
                                {MA_RTC + RTCO_MIN,           1, T_DATAPTR,                (uint16_t) &rtc_min,               NULL,     NULL},            /* RTC minute */
                                {MA_RTC + RTCO_SEC,           1, T_DATAPTR,                (uint16_t) &rtc_sec,               NULL,     NULL},            /* RTC seconds */
                                {MA_RTC + RTCO_CS,            1, T_DATAPTR,                (uint16_t) &rtc_cs,                NULL,     NULL},            /* RTC centiseconds */
                                {MA_RTC + RTCO_ENABLE,        1, T_DATAPTR,                (uint16_t) &rtc_active,            NULL,     NULL},            /* Enable RTC */
//                               |                            |                            |                                  |         |
                                {MA_SYSREG + SREGO_RESTART,   1, T_EXECUTE,                0,                                 NULL,     restart},         /* Write MAGIC_NUMBER to restart device */
                                {MA_SYSREG + SREGO_CLR_CNT,   1, T_EXECUTE,                0,                                 NULL,     clear_counters},  /* Write MAGIC_NUMBER to clear error counters */
                                {MA_SYSREG + SREGO_RESDEVID,  1, T_EXECUTE,                0,                                 NULL,     reset_deviceid},  /* Reset device ID in EEPROM to recover from multiple devices with same ID */
                                {MA_SYSREG + SREGO_STARTBOOT, 1, T_EXECUTE,                0,                                 NULL,     start_bootloader},/* Start bootloader */
//                               |                            |                            |                                  |         |
                                {MA_STATUS + STRO_MCUSR,      1, T_DATAPTR,                (uint16_t) &reset_flags,           NULL,     NULL},            /* Read MCU Status Register */
                                {MA_STATUS + STRO_RESET,      1, T_DATAPTR | T_WORD,       (uint16_t) &reset_count,           NULL,     NULL},            /* Read total number of resets since last power-on reset */
                                {MA_STATUS + STRO_WDTR,       1, T_DATAPTR | T_WORD,       (uint16_t) &wdt_reset_count,       NULL,     NULL},            /* Read number of watchdog timer resets since last power-on reset */
                                {MA_STATUS + STRO_SR,         1, T_DATAPTR | T_WORD,       (uint16_t) &software_reset_count,  NULL,     NULL},            /* Read number of software resets since last power-on reset */
                                {MA_STATUS + STRO_EXTR,       1, T_DATAPTR | T_WORD,       (uint16_t) &ext_reset_count,       NULL,     NULL},            /* Read number of external resets since last power-on reset */
                                {MA_STATUS + STRO_BOR,        1, T_DATAPTR | T_WORD,       (uint16_t) &bo_reset_count,        NULL,     NULL},            /* Read number of brown-out resets since last power-on reset */
                                {MA_STATUS + STRO_TMSGH,      1, T_DATAPTR | T_WORD,       (uint16_t) &total_message_count_h, NULL,     NULL},            /* Read total number of error free messages since last power-on reset (higher) */
                                {MA_STATUS + STRO_TMSGL,      1, T_DATAPTR | T_WORD,       (uint16_t) &total_message_count_l, NULL,     NULL},            /* Read total number of error free messages since last power-on reset (lower) */
                                {MA_STATUS + STRO_MSGH,       1, T_DATAPTR | T_WORD,       (uint16_t) &message_count_h,       NULL,     NULL},            /* Read number of messages for this module since last power-on reset (higher) */
                                {MA_STATUS + STRO_MSGL,       1, T_DATAPTR | T_WORD,       (uint16_t) &message_count_l,       NULL,     NULL},            /* Read number of messages for this module since last power-on reset (lower) */
                                {MA_STATUS + STRO_CRCEH,      1, T_DATAPTR | T_WORD,       (uint16_t) &crc_error_count_h,     NULL,     NULL},            /* Read number of CRC errors since last power-on reset (higher) */
                                {MA_STATUS + STRO_CRCEL,      1, T_DATAPTR | T_WORD,       (uint16_t) &crc_error_count_l,     NULL,     NULL},            /* Read number of CRC errors since last power-on reset (lower) */
                                {MA_STATUS + STRO_COMH,       1, T_DATAPTR | T_WORD,       (uint16_t) &com_error_count_h,     NULL,     NULL},            /* Read number of communication errors since last power-on reset (higher) */
                                {MA_STATUS + STRO_COML,       1, T_DATAPTR | T_WORD,       (uint16_t) &com_error_count_l,     NULL,     NULL},            /* Read number of communication errors since last power-on reset (lower) */
                                {MA_STATUS + STRO_RXCNTH,     1, T_DATAPTR | T_WORD,       (uint16_t) &com_rx_count_h,        NULL,     NULL},            /* Read number of received bytes since last power-on reset (higher) */
                                {MA_STATUS + STRO_RXCNTL,     1, T_DATAPTR | T_WORD,       (uint16_t) &com_rx_count_l,        NULL,     NULL},            /* Read number of received bytes since last power-on reset (lower) */
                                {MA_STATUS + STRO_TXCNTH,     1, T_DATAPTR | T_WORD,       (uint16_t) &com_tx_count_h,        NULL,     NULL},            /* Read number of transmitted bytes since last power-on reset (higher) */
                                {MA_STATUS + STRO_TXCNTL,     1, T_DATAPTR | T_WORD,       (uint16_t) &com_tx_count_l,        NULL,     NULL},            /* Read number of transmitted bytes since last power-on reset (lower) */
                                {MA_STATUS + STRO_ONTIMEH,    1, T_DATAPTR | T_WORD,       (uint16_t) &on_time_h,             NULL,     NULL},            /* Power-on time in centi-seconds (higher) */
                                {MA_STATUS + STRO_ONTIMEM,    1, T_DATAPTR | T_WORD,       (uint16_t) &on_time_m,             NULL,     NULL},            /* Power-on time in centi-seconds (middle) */
                                {MA_STATUS + STRO_ONTIMEL,    1, T_DATAPTR | T_WORD,       (uint16_t) &on_time_l,             NULL,     NULL},            /* Power-on time in centi-seconds (lower) */
//                               |                            |                            |                                  |         |
                                {MA_TELEM +  TREGO_TEMP,      1, T_DATAPTR | T_RO | T_WORD,(uint16_t) &temperature,           NULL,     NULL},            /* Read temperature */
                                {MA_TELEM +  TREGO_VSUPP,     1, T_DATAPTR | T_RO | T_WORD,(uint16_t) &vin_voltage,           NULL,     NULL},            /* Read supply voltage Vin */
//                               |                            |                            |                                  |         |
                                {MA_FUNC + FCRO_RELAYS_I,     1, T_EXECUTE,                0,                                 get_relay,set_relay},       /* Relays for inductors */
                                {MA_FUNC + FCRO_RELAYS_C,     1, T_EXECUTE,                0,                                 get_relay,set_relay},       /* Relays for capacitors */
//                               |                            |                            |                                  |         |
                                {MA_PARMS, sizeof(TCXO)/2,       T_EEPPTR | T_WORD,        (uint16_t) &tcxo_parms,            NULL,     NULL},            /* read/write EEPROM parameters */
//                               |                            |                            |                                  |         |
//                                {MA_TEST,                   16, T_DATAPTR | T_WORD,       (uint16_t) &testdata,              NULL,     NULL},            /* read/write RAM multiple */
        } ;

const PROGMEM MODBUS_DESCRIPTOR modbus_desc = {
        (PROGMEM MODBUS_REGISTER *) Input_Register,   sizeof(Input_Register)/sizeof(Input_Register[0]),
        (PROGMEM MODBUS_REGISTER *) Holding_Register, sizeof(Holding_Register)/sizeof(Holding_Register[0]),
        (PROGMEM MODBUS_REGISTER *) Coils,            sizeof(Coils)/sizeof(Coils[0]),
        (PROGMEM MODBUS_REGISTER *) File_Records,     sizeof(File_Records)/sizeof(File_Records[0])
        } ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Inline functions                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/
__inline static void eeprom_write_byte_cond(uint8_t  *addr, uint8_t  data) {if (eeprom_read_byte(addr) != data) eeprom_write_byte(addr, data) ; }
__inline static void eeprom_write_word_cond(uint16_t *addr, uint16_t data) {if (eeprom_read_word(addr) != data) eeprom_write_word(addr, data) ; }

#if (MODULE_TYPE==MT_TUNER_V10)
__inline static uint8_t sort_i_relays(uint8_t d)  {return((d&0x20)>>2 | (d&0x10)>>3 | (d&0x08)<<1 | (d&0x04)>>0 | (d&0x02)<<4 | (d&0x01)>>0) ; } /* shift bits into proper position */
__inline static uint8_t sort_c_relays(uint8_t d)  {return(d) ;} /* shift bits into proper position */
#endif
#if (MODULE_TYPE==MT_TUNER_V12)
__inline static uint8_t sort_i_relays(uint8_t d)  {return((d&0x80)>>0 | (d&0x40)>>0 | (d&0x20)>>0 | (d&0x10)>>0 | (d&0x08)<<0 | (d&0x04)>>1 | (d&0x02)<<1 | (d&0x01)>>0) ; } /* shift bits into proper position */
__inline static uint8_t sort_c_relays(uint8_t d)  {return(d) ;} /* shift bits into proper position */
#endif

__inline static void relay_lowside_on(uint8_t relay_nr) {PORT_LOWSIDE |= 1<<relay_nr ;}
__inline static void relay_lowside_off(uint8_t relay_nr) {PORT_LOWSIDE &= ~(1<<relay_nr) ; }
__inline static void relay_lowside_all_off(void) {PORT_LOWSIDE &= ~PIN_LOWDRIVERS ; }
__inline static void relay_highside_all_off(void) {PORT_HIGHSIDE &= ~(PIN_CRELAY_ON | PIN_CRELAY_OFF | PIN_IRELAY_ON | PIN_IRELAY_OFF) ;}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Main                                                                                                            |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| MAIN program.                                                                                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
int main(void)
{ /* executed on coldstart */

   initial_things() ;

#ifdef SIMULATION
/* some test code, may be removed later */
#define NewTest

#ifdef NewTest
   init_CPU() ; /* initialize CPU, turn global interrupts on */
   /* Initialise EEPROM with default values, for debugging */
   mcb_eepinit(&eemcb[0], &default_mcb) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x03 ; /* function code */
   rxbuf.buffer[2] = 0x0c ; /* address high */
   rxbuf.buffer[3] = 0x27 ; /* address low */
   rxbuf.buffer[4] = 0 ; /* count high */
   rxbuf.buffer[5] = 1 ; /* count low */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x06 ; /* function code */
   rxbuf.buffer[2] = (RELAYS_I)>>8 ; /* address high */
   rxbuf.buffer[3] = (RELAYS_I)&0xff ; /* address low */
   rxbuf.buffer[4] = 0xa3 ; /* value */
   rxbuf.buffer[5] = 0x15 ; /* value */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x05 ; /* function code */
   rxbuf.buffer[2] = (VIRTUAL_ADDRESS_PORTS+17)>>8 ; /* address high */
   rxbuf.buffer[3] = (VIRTUAL_ADDRESS_PORTS+17)&0xff ; /* address low */
   rxbuf.buffer[4] = 0xff ; /* coil on */
   rxbuf.buffer[5] = 0 ; /* always 0 */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x03 ; /* function code */
   rxbuf.buffer[2] = 0x0b ; /* address high */
   rxbuf.buffer[3] = 0xb8 ; /* address low */
   rxbuf.buffer[4] = 0 ; /* count high */
   rxbuf.buffer[5] = 1 ; /* count low */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x03 ; /* function code */
   rxbuf.buffer[2] = VA_MCB>>8 ; /* address high */
   rxbuf.buffer[3] = VA_MCB&0xff ; /* address low */
   rxbuf.buffer[4] = 0 ; /* count high */
   rxbuf.buffer[5] = 5 ; /* count low */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x10 ; /* function code */
   rxbuf.buffer[2] = (VA_RTC+ 0)>>8 ; /* address high */
   rxbuf.buffer[3] = (VA_RTC+ 0)&0xff ; /* address low */
   rxbuf.buffer[4] = 0x00 ; /* file number high */
   rxbuf.buffer[5] = 0x01 ; /* file number low */
   rxbuf.buffer[6] = 2 ; /* Byte count */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x14 ; /* function code */
   rxbuf.buffer[2] = 0x07 ; /* byte count */
   rxbuf.buffer[3] = 0x06 ; /* reference type */
   rxbuf.buffer[4] = 0x00 ; /* file number high */
   rxbuf.buffer[5] = 0x01 ; /* file number low */
   rxbuf.buffer[6] = (VA_FR + 0)>>8 ; /* record number high */
   rxbuf.buffer[7] = (VA_FR + 0)&0xff ; /* record number low */
   rxbuf.buffer[8] = 0 ;    /* length high */
   rxbuf.buffer[9] = 30  ;  /* length low */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x06 ; /* function code */
   rxbuf.buffer[2] = VA_RTC>>8 ; /* address high */
   rxbuf.buffer[3] = VA_RTC&0xff ; /* address low */
   rxbuf.buffer[4] = 0x4d ; /* data high */
   rxbuf.buffer[5] = 0x9e ; /* data low */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x06 ; /* function code */
   rxbuf.buffer[2] = (VA_STATUS+2)>>8 ; /* address high */
   rxbuf.buffer[3] = (VA_STATUS+2)&0xff ; /* address low */
   rxbuf.buffer[4] = 0x7e ; /* data high */
   rxbuf.buffer[5] = 0x12 ; /* data low */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x01 ; /* function code */
   rxbuf.buffer[2] = VIRTUAL_ADDRESS_PORTS>>8 ; /* address high */
   rxbuf.buffer[3] = VIRTUAL_ADDRESS_PORTS&0xff ; /* address low */
   rxbuf.buffer[4] = 0 ; /* count high */
   rxbuf.buffer[5] = 12 ; /* count low */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x04 ; /* function code */
   rxbuf.buffer[2] = VIRTUAL_ADDRESS_IR>>8 ; /* address high */
   rxbuf.buffer[3] = VIRTUAL_ADDRESS_IR&0xff ; /* address low */
   rxbuf.buffer[4] = 0 ; /* count high */
   rxbuf.buffer[5] = 6 ; /* count low */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;

   rxbuf.buffer[0] = MODBUS_ADDR ; /* slave address */
   rxbuf.buffer[1] = 0x04 ; /* function code */
   rxbuf.buffer[2] = VA_EEP>>8 ; /* address high */
   rxbuf.buffer[3] = VA_EEP&0xff ; /* address low */
   rxbuf.buffer[4] = 0 ; /* count high */
   rxbuf.buffer[5] = 32 ; /* count low */
   modbus_message_handler(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;
#endif
#endif

   init_CPU() ; /* initialize CPU, turn global interrupts on */

   wdt_on() ; /* switch watchdog timer on */

   get_supported_baudrates() ;

   listen_only = false ;
   restart_communications = false ;

   for (;;)
   {
/* Setup serial receiver and wait for message */
      receive(&rxbuf) ;
      wait_for_rx(&rxbuf) ; /* wait for received message */

      switch(rxbuf.status)
      { /* reception completed. Check for error, otherwise check address and prepare response */
         case RXSTATE_COMPLETED:
            modbus_message_received(&rxbuf, &txbuf, (MODBUS_DESCRIPTOR *) &modbus_desc) ;
         break ;

         case RXSTATE_ERROR:
         case RXSTATE_BUFFEROVERFLOW :
         /* error, just increment error counter */
            com_error_count_l++ ;
            if (com_error_count_l == 0) com_error_count_h++ ;
         break ;

         default: /* ignore other cases for now */
         break ;
      }
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void wait_for_rx_user(void)                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| User specific wait tasks.                                                                                             |
|                                                                                                                       |
| This function executes regularly while waiting for a bus transaction. Interrupts shall not be disabled.               |
| Bus messages may be lost, if execution takes too long.                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void wait_for_rx_user(void)
{
   switch (relay_state)
   {
      case 0: /* inactive, do nothing */

      break ;

      case 1: /* state machine for the inductor relays is to be started */
         next_relay = 0 ; /* initialize variables */
         relay_on_highdriver = PIN_IRELAY_ON ;
         relay_off_highdriver = PIN_IRELAY_OFF ;
         relays_target = sort_i_relays(relays_i) ;
         relay_highside_all_off() ; /* ensure all drivers are off */
         relay_lowside_all_off() ;
         relay_state = 3 ; /* proceed to first state */
         relay_timer.timer = 1 ;
         relay_timer.active = true ;
      break ;

      case 2: /* state machine for the capacitor relays is to be started */
         next_relay = 0 ; /* initialize variables */
         relay_on_highdriver = PIN_CRELAY_ON ;
         relay_off_highdriver = PIN_CRELAY_OFF ;
         relays_target = sort_c_relays(relays_c) ;
         relay_highside_all_off() ; /* ensure all drivers are off */
         relay_lowside_all_off() ;
         relay_state = 3 ; /* proceed to first state */
         relay_timer.timer = 1 ;
         relay_timer.active = true ;
      break ;

      case 3:
      case 5:
      case 7:
      case 9:
      case 11:
      case 13:
#if (MODULE_TYPE==MT_TUNER_V12)
      case 15:
      case 17:
#endif
         if (!relay_timer.active) /* switch a relay on or off, start set delay time */
         { /* timer has expired, switch on selected high side and low side drivers */
            PORTB |= ((relays_target & 1<<next_relay) != 0)?relay_on_highdriver:relay_off_highdriver ;
            relay_lowside_on(next_relay) ;
            next_relay++ ;
            relay_state++ ; /* proceed to next state */
            relay_timer.timer = RELAY_SET_TIME ;
            relay_timer.active = true ;
         }
      break ;

      case 4:
      case 6:
      case 8:
      case 10:
      case 12:
      case 14:
#if (MODULE_TYPE==MT_TUNER_V12)
      case 16:
      case 18:
#endif
         if (!relay_timer.active) /* switch driver off and start recovery delay */
         { /* timer has expired, switch off all high side drivers */
            relay_highside_all_off() ; /* ensure all drivers are off */
            relay_lowside_all_off() ;
            relay_state++ ; /* proceed to next state */
            relay_timer.timer = RELAY_RECOVERY_TIME ;
            relay_timer.active = true ;
         }
      break ;

      default: /* all undefined states go back idle */
         if (next_relay_state != 0) relay_state = next_relay_state ; /* append another sequence when completed */
         else relay_state = 0 ;
         next_relay_state = 0 ;
      break ;
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       readcoil                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| This is a callback function for the modbus library. It reads one single coil.                                         |
|                                                                                                                       |
| Returns true if successful, false on error                                                                            |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool readcoil(uint16_t addr, uint16_t *data)
{
/*
Port mapping is as follows:
MA_PORTS+ 0.. 7: PA0..PA7
MA_PORTS+ 8..15: PB0..PB7
MA_PORTS+16..23: PC0..PC7
MA_PORTS+24..31: PD0..PD7
*/

#ifdef DEBUGCODE

   if ((addr >= (MA_PORTS+ 0)) && (addr < (MA_PORTS+8)))
   {
      *data = ((PINA >> (addr - (MA_PORTS+ 0))) & 1) ; /* Port A access */
      return(true) ;
   }
   if ((addr >= (MA_PORTS+ 8)) && (addr < (MA_PORTS+16)))
   {
      *data = ((PINB >> (addr - (MA_PORTS+ 8))) & 1) ; /* Port B access */
      return(true) ;
   }
   if ((addr >= (MA_PORTS+16)) && (addr < (MA_PORTS+24)))
   {
      *data = ((PINC >> (addr - (MA_PORTS+16))) & 1) ; /* Port C access */
      return(true) ;
   }
   if ((addr >= (MA_PORTS+24)) && (addr < (MA_PORTS+32)))
   {
      *data = ((PIND >> (addr - (MA_PORTS+24))) & 1) ; /* Port D access */
      return(true) ;
   }

#endif

   return(false) ; /* should not happen if DEBUGCODE is defined */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       writecoil                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| This is a callback function for the modbus library. It writes one single coil.                                        |
|                                                                                                                       |
| Returns true if successful, false on error                                                                            |
|                                                                                                                       |
| *data will be 0x0000 if coil is 0, 0xff00 if coil is 1.                                                               |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool writecoil(uint16_t addr, uint16_t *data)
{
/*
Port mapping is as follows:
MA_PORTS+ 0.. 7: PA0..PA7
MA_PORTS+ 8..15: PB0..PB7
MA_PORTS+16..23: PC0..PC7
MA_PORTS+24..31: PD0..PD7
*/

#ifdef DEBUGCODE
uint8_t mask ;

   if ((addr >= (MA_PORTS + 0)) && (addr < (MA_PORTS + 8)))
   { /* Port A */
      mask = (1<<(addr-(MA_PORTS+0))) & PORTA_DIR ;
      if (*data != 0) PORTA |=  mask ; /* set selected output port */
      else            PORTA &= ~mask ; /* clear selected port */
      _delay_loop_1 (10) ;             /* small delay to allow port to switch */
      *data = (((PINA & mask)==0)?0x0000:0xff00) ;
      return(true) ;
   }

   if ((addr >= (MA_PORTS + 8)) && (addr < (MA_PORTS +16)))
   { /* Port B */
      mask = (1<<(addr-(MA_PORTS+8))) & PORTB_DIR ;
      if (*data != 0) PORTB |=  mask ; /* set selected output port */
      else            PORTB &= ~mask ; /* clear selected port */
      _delay_loop_1 (10) ;             /* small delay to allow port to switch */
      *data = (((PINB & mask)==0)?0x0000:0xff00) ;
      return(true) ;
   }

   if ((addr >= (MA_PORTS +16)) && (addr < (MA_PORTS +24)))
   { /* Port C */
      mask = (1<<(addr-(MA_PORTS+16))) & PORTC_DIR ;
      if (*data != 0) PORTC |=  mask ; /* set selected output port */
      else            PORTC &= ~mask ; /* clear selected port */
      _delay_loop_1 (10) ;             /* small delay to allow port to switch */
      *data = (((PINC & mask)==0)?0x0000:0xff00) ;
      return(true) ;
   }

   if ((addr >= (MA_PORTS +24)) && (addr < (MA_PORTS +32)))
   { /* Port D */
      mask = (1<<(addr-(MA_PORTS+24))) & PORTD_DIR ;
      if (*data != 0) PORTD |=  mask ; /* set selected output port */
      else            PORTD &= ~mask ; /* clear selected port */
      _delay_loop_1 (10) ;             /* small delay to allow port to switch */
      *data = (((PIND & mask)==0)?0x0000:0xff00) ;
      return(true) ;
   }

#endif

   return(false) ; /* should not happen if DEBUGCODE is defined */

}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       set_relay                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Set relays                                                                                                            |
|                                                                                                                       |
| This is a callback function for write holding registers                                                               |
|                                                                                                                       |
| Returns false if the address or data is illegal, true on success.                                                     |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool set_relay(uint16_t addr, uint16_t *data)
{
   switch(addr)
   {
      case (MA_FUNC + FCRO_RELAYS_I): /* set relays for inductors */
         relays_i = *data & PIN_LOWDRIVERS ;
         relays_i_valid = true ;
         if (relay_state == 0)
            relay_state = 1 ; /* start state machine to switch the relays */
         else
            next_relay_state = 1 ;
         return(true) ;
      break ;

      case (MA_FUNC + FCRO_RELAYS_C): /* set relays for capacitors */
         relays_c = *data & PIN_LOWDRIVERS ;
         relays_c_valid = true ;
         if (relay_state == 0)
            relay_state = 2 ; /* start state machine to switch the relays */
         else
            next_relay_state = 2 ;
         return(true) ;
      break ;

      default:
         return(false) ; /* error */
      break ;
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       get_relay                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Get relays                                                                                                            |
|                                                                                                                       |
| This is a callback function for read holding registers                                                                |
|                                                                                                                       |
| Returns false if the address is illegal, true on success.                                                             |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool get_relay(uint16_t addr, uint16_t *data)
{
   switch(addr)
   {
      case (MA_FUNC + FCRO_RELAYS_I): /* get state of relays for inductors */
         if ((data != NULL) && (relays_i_valid))
         {
            *data = relays_i ;
            return(true) ;
         }
      break ;

      case (MA_FUNC + FCRO_RELAYS_C): /* get state of relays for capacitors */
         if ((data != NULL) && (relays_c_valid))
         {
            *data = relays_c ;
            return(true) ;
         }
      break ;
   }

   return(false) ; /* error */
}

